#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Generator
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.rds import RdsHook
from airflow.providers.common.compat.sdk import AirflowException, AirflowNotFoundException

if TYPE_CHECKING:
    from mypy_boto3_rds.type_defs import DBSnapshotTypeDef


@pytest.fixture
def rds_hook() -> Generator[RdsHook, None, None]:
    """Returns an RdsHook whose underlying connection is mocked with moto"""
    with mock_aws():
        yield RdsHook(aws_conn_id="aws_default", region_name="us-east-1")


@pytest.fixture
def db_instance_id(rds_hook: RdsHook) -> str:
    """Creates an RDS DB instance and returns its id"""
    response = rds_hook.conn.create_db_instance(
        DBInstanceIdentifier="testrdshook-db-instance",
        DBInstanceClass="db.t4g.micro",
        Engine="postgres",
        AllocatedStorage=20,
        MasterUsername="testrdshook",
        MasterUserPassword="testrdshook",
    )
    return response["DBInstance"]["DBInstanceIdentifier"]


@pytest.fixture
def db_cluster_id(rds_hook: RdsHook) -> str:
    """Creates an RDS DB cluster and returns its id"""
    response = rds_hook.conn.create_db_cluster(
        DBClusterIdentifier="testrdshook-db-cluster",
        Engine="postgres",
        MasterUsername="testrdshook",
        MasterUserPassword="testrdshook",
        DBClusterInstanceClass="db.t4g.micro",
        AllocatedStorage=20,
    )
    return response["DBCluster"]["DBClusterIdentifier"]


@pytest.fixture
def db_snapshot(rds_hook: RdsHook, db_instance_id: str) -> DBSnapshotTypeDef:
    """
    Creates a mock DB instance snapshot and returns the DBSnapshot dict from the boto response object.
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/rds.html#RDS.Client.create_db_snapshot
    """
    response = rds_hook.conn.create_db_snapshot(
        DBSnapshotIdentifier="testrdshook-db-instance-snapshot", DBInstanceIdentifier=db_instance_id
    )
    return response["DBSnapshot"]


@pytest.fixture
def db_snapshot_id(db_snapshot: dict) -> str:
    return db_snapshot["DBSnapshotIdentifier"]


@pytest.fixture
def db_snapshot_arn(db_snapshot: dict) -> str:
    return db_snapshot["DBSnapshotArn"]


@pytest.fixture
def db_cluster_snapshot(rds_hook: RdsHook, db_cluster_id: str):
    """
    Creates a mock DB cluster snapshot and returns the DBClusterSnapshot dict from the boto response object.
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/rds.html#RDS.Client.create_db_cluster_snapshot
    """
    response = rds_hook.conn.create_db_cluster_snapshot(
        DBClusterSnapshotIdentifier="testrdshook-db-cluster-snapshot", DBClusterIdentifier=db_cluster_id
    )
    return response["DBClusterSnapshot"]


@pytest.fixture
def db_cluster_snapshot_id(db_cluster_snapshot) -> str:
    return db_cluster_snapshot["DBClusterSnapshotIdentifier"]


@pytest.fixture
def export_task_id(rds_hook: RdsHook, db_snapshot_arn: str) -> str:
    response = rds_hook.conn.start_export_task(
        ExportTaskIdentifier="testrdshook-export-task",
        SourceArn=db_snapshot_arn,
        S3BucketName="test",
        IamRoleArn="test",
        KmsKeyId="test",
    )
    return response["ExportTaskIdentifier"]


@pytest.fixture
def event_subscription_name(rds_hook: RdsHook, db_instance_id: str) -> str:
    """Creates an mock RDS event subscription and returns its name"""
    response = rds_hook.conn.create_event_subscription(
        SubscriptionName="testrdshook-event-subscription",
        SnsTopicArn="test",
        SourceType="db-instance",
        SourceIds=[db_instance_id],
        Enabled=True,
    )
    return response["EventSubscription"]["CustSubscriptionId"]


class TestRdsHook:
    # For testing, set the delay between status checks to 0 so that we aren't sleeping during tests,
    # and max_attempts to 1 so that we don't retry unless required.
    waiter_args = {"check_interval": 0, "max_attempts": 1}

    def test_conn_attribute(self):
        hook = RdsHook(aws_conn_id="aws_default", region_name="us-east-1")
        assert hasattr(hook, "conn")
        assert hook.conn.__class__.__name__ == "RDS"
        conn = hook.conn
        assert conn is hook.conn  # Cached property
        assert conn is hook.get_conn()  # Same object as returned by `conn` property

    def test_get_db_instance_state(self, rds_hook: RdsHook, db_instance_id: str):
        response = rds_hook.conn.describe_db_instances(DBInstanceIdentifier=db_instance_id)
        state_expected = response["DBInstances"][0]["DBInstanceStatus"]
        state_actual = rds_hook.get_db_instance_state(db_instance_id)
        assert state_actual == state_expected

    def test_wait_for_db_instance_state_boto_waiters(self, rds_hook: RdsHook, db_instance_id: str):
        """Checks that the DB instance waiter uses AWS boto waiters where possible"""
        for state in ("available", "deleted", "stopped"):
            with patch.object(rds_hook.conn, "get_waiter") as mock:
                rds_hook.wait_for_db_instance_state(db_instance_id, target_state=state, **self.waiter_args)
                mock.assert_called_once_with(f"db_instance_{state}")
                mock.return_value.wait.assert_called_once_with(
                    DBInstanceIdentifier=db_instance_id,
                    WaiterConfig={
                        "MaxAttempts": self.waiter_args["max_attempts"],
                    },
                )

    def test_get_db_cluster_state(self, rds_hook: RdsHook, db_cluster_id: str):
        response = rds_hook.conn.describe_db_clusters(DBClusterIdentifier=db_cluster_id)
        state_expected = response["DBClusters"][0]["Status"]
        state_actual = rds_hook.get_db_cluster_state(db_cluster_id)
        assert state_actual == state_expected

    def test_wait_for_db_cluster_state_boto_waiters(self, rds_hook: RdsHook, db_cluster_id: str):
        """Checks that the DB cluster waiter uses AWS boto waiters where possible"""
        for state in ("available", "deleted", "stopped"):
            with patch.object(rds_hook.conn, "get_waiter") as mock:
                rds_hook.wait_for_db_cluster_state(db_cluster_id, target_state=state, **self.waiter_args)
                mock.assert_called_once_with(f"db_cluster_{state}")
                mock.return_value.wait.assert_called_once_with(
                    DBClusterIdentifier=db_cluster_id,
                    WaiterConfig={
                        "Delay": self.waiter_args["check_interval"],
                        "MaxAttempts": self.waiter_args["max_attempts"],
                    },
                )

    def test_get_db_snapshot_state(self, rds_hook: RdsHook, db_snapshot_id: str):
        response = rds_hook.conn.describe_db_snapshots(DBSnapshotIdentifier=db_snapshot_id)
        state_expected = response["DBSnapshots"][0]["Status"]
        state_actual = rds_hook.get_db_snapshot_state(db_snapshot_id)
        assert state_actual == state_expected

    def test_get_db_snapshot_state_not_found(self, rds_hook: RdsHook):
        with pytest.raises(AirflowNotFoundException):
            rds_hook.get_db_snapshot_state("does_not_exist")

    def test_wait_for_db_snapshot_state_boto_waiters(self, rds_hook: RdsHook, db_snapshot_id: str):
        """Checks that the DB snapshot waiter uses AWS boto waiters where possible"""
        for state in ("available", "deleted", "completed"):
            with patch.object(rds_hook.conn, "get_waiter") as mock:
                rds_hook.wait_for_db_snapshot_state(db_snapshot_id, target_state=state, **self.waiter_args)
                mock.assert_called_once_with(f"db_snapshot_{state}")
                mock.return_value.wait.assert_called_once_with(
                    DBSnapshotIdentifier=db_snapshot_id,
                    WaiterConfig={
                        "Delay": self.waiter_args["check_interval"],
                        "MaxAttempts": self.waiter_args["max_attempts"],
                    },
                )

    def test_wait_for_db_snapshot_state_custom_waiter(self, rds_hook: RdsHook, db_snapshot_id: str):
        """Checks that the DB snapshot waiter uses custom wait logic when AWS boto waiters aren't available"""
        with patch.object(rds_hook, "_wait_for_state") as mock:
            rds_hook.wait_for_db_snapshot_state(db_snapshot_id, target_state="canceled", **self.waiter_args)
            mock.assert_called_once()

        with patch.object(rds_hook, "get_db_snapshot_state", return_value="canceled") as mock:
            rds_hook.wait_for_db_snapshot_state(db_snapshot_id, target_state="canceled", **self.waiter_args)
            mock.assert_called_once_with(db_snapshot_id)

    def test_get_db_cluster_snapshot_state(self, rds_hook: RdsHook, db_cluster_snapshot_id: str):
        response = rds_hook.conn.describe_db_cluster_snapshots(
            DBClusterSnapshotIdentifier=db_cluster_snapshot_id
        )
        state_expected = response["DBClusterSnapshots"][0]["Status"]
        state_actual = rds_hook.get_db_cluster_snapshot_state(db_cluster_snapshot_id)
        assert state_actual == state_expected

    def test_get_db_cluster_snapshot_state_not_found(self, rds_hook: RdsHook):
        with pytest.raises(AirflowNotFoundException):
            rds_hook.get_db_cluster_snapshot_state("does_not_exist")

    def test_wait_for_db_cluster_snapshot_state_boto_waiters(
        self, rds_hook: RdsHook, db_cluster_snapshot_id: str
    ):
        """Checks that the DB cluster snapshot waiter uses AWS boto waiters where possible"""
        for state in ("available", "deleted"):
            with patch.object(rds_hook.conn, "get_waiter") as mock:
                rds_hook.wait_for_db_cluster_snapshot_state(
                    db_cluster_snapshot_id, target_state=state, **self.waiter_args
                )
                mock.assert_called_once_with(f"db_cluster_snapshot_{state}")
                mock.return_value.wait.assert_called_once_with(
                    DBClusterSnapshotIdentifier=db_cluster_snapshot_id,
                    WaiterConfig={
                        "Delay": self.waiter_args["check_interval"],
                        "MaxAttempts": self.waiter_args["max_attempts"],
                    },
                )

    def test_wait_for_db_cluster_snapshot_state_custom_waiter(
        self, rds_hook: RdsHook, db_cluster_snapshot_id: str
    ):
        """
        Checks that the DB cluster snapshot waiter uses custom wait logic when AWS boto waiters
        aren't available
        """
        with patch.object(rds_hook, "_wait_for_state") as mock:
            rds_hook.wait_for_db_cluster_snapshot_state(
                db_cluster_snapshot_id, target_state="canceled", **self.waiter_args
            )
            mock.assert_called_once()

        with patch.object(rds_hook, "get_db_cluster_snapshot_state", return_value="canceled") as mock:
            rds_hook.wait_for_db_cluster_snapshot_state(
                db_cluster_snapshot_id, target_state="canceled", **self.waiter_args
            )
            mock.assert_called_once_with(db_cluster_snapshot_id)

    def test_get_export_task_state(self, rds_hook: RdsHook, export_task_id: str):
        response = rds_hook.conn.describe_export_tasks(ExportTaskIdentifier=export_task_id)
        state_expected = response["ExportTasks"][0]["Status"]
        state_actual = rds_hook.get_export_task_state(export_task_id)
        assert state_actual == state_expected

    def test_get_export_task_state_not_found(self, rds_hook: RdsHook):
        with pytest.raises(AirflowNotFoundException):
            rds_hook.get_export_task_state("does_not_exist")

    def test_wait_for_export_task_state(self, rds_hook: RdsHook, export_task_id: str):
        """
        Checks that the export task waiter uses custom wait logic (no boto waiters exist for this resource)
        """
        with patch.object(rds_hook, "_wait_for_state") as mock:
            rds_hook.wait_for_export_task_state(export_task_id, target_state="complete", **self.waiter_args)
            mock.assert_called_once()

        with patch.object(rds_hook, "get_export_task_state", return_value="complete") as mock:
            rds_hook.wait_for_export_task_state(export_task_id, target_state="complete", **self.waiter_args)
            mock.assert_called_once_with(export_task_id)

    def test_get_event_subscription_state(self, rds_hook: RdsHook, event_subscription_name: str):
        response = rds_hook.conn.describe_event_subscriptions(SubscriptionName=event_subscription_name)
        state_expected = response["EventSubscriptionsList"][0]["Status"]
        state_actual = rds_hook.get_event_subscription_state(event_subscription_name)
        assert state_actual == state_expected

    def test_get_event_subscription_state_not_found(self, rds_hook: RdsHook):
        with pytest.raises(AirflowNotFoundException):
            rds_hook.get_event_subscription_state("does_not_exist")

    def test_wait_for_event_subscription_state(self, rds_hook: RdsHook, event_subscription_name: str):
        """
        Checks that the event subscription waiter uses custom wait logic (no boto waiters
        exist for this resource)
        """
        with patch.object(rds_hook, "_wait_for_state") as mock:
            rds_hook.wait_for_event_subscription_state(
                event_subscription_name, target_state="active", **self.waiter_args
            )
            mock.assert_called_once()

        with patch.object(rds_hook, "get_event_subscription_state", return_value="active") as mock:
            rds_hook.wait_for_event_subscription_state(
                event_subscription_name, target_state="active", **self.waiter_args
            )
            mock.assert_called_once_with(event_subscription_name)

    def test_wait_for_state(self, rds_hook: RdsHook):
        def poke():
            return "foo"

        with pytest.raises(AirflowException, match="Max attempts exceeded"):
            with patch("airflow.providers.amazon.aws.hooks.rds.time.sleep") as mock:
                rds_hook._wait_for_state(poke, target_state="bar", check_interval=0, max_attempts=2)
        # This next line should exist outside of the pytest.raises() context manager or else it won't
        # get executed
        mock.assert_called_once_with(0)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
import requests
from botocore.exceptions import ClientError
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.mwaa import MwaaHook

ENV_NAME = "test_env"
PATH = "/dags/test_dag/dagRuns"
METHOD = "POST"
BODY: dict = {"conf": {}}
QUERY_PARAMS = {"limit": 30}
HOSTNAME = "example.com"


class TestMwaaHook:
    @pytest.fixture
    def mock_conn(self):
        with mock.patch.object(MwaaHook, "conn") as m:
            yield m

    def setup_method(self):
        self.hook = MwaaHook()

    def test_init(self):
        assert self.hook.client_type == "mwaa"

    @mock_aws
    def test_get_conn(self):
        assert self.hook.conn is not None

    @pytest.mark.parametrize(
        "body",
        [
            pytest.param(None, id="no_body"),
            pytest.param(BODY, id="non_empty_body"),
        ],
    )
    def test_invoke_rest_api_success(self, body, mock_conn, example_responses):
        boto_invoke_mock = mock.MagicMock(return_value=example_responses["success"])
        mock_conn.invoke_rest_api = boto_invoke_mock

        retval = self.hook.invoke_rest_api(
            env_name=ENV_NAME, path=PATH, method=METHOD, body=body, query_params=QUERY_PARAMS
        )
        kwargs_to_assert = {
            "Name": ENV_NAME,
            "Path": PATH,
            "Method": METHOD,
            "Body": body if body else {},
            "QueryParameters": QUERY_PARAMS,
        }
        boto_invoke_mock.assert_called_once_with(**kwargs_to_assert)
        mock_conn.create_web_login_token.assert_not_called()
        assert retval == {k: v for k, v in example_responses["success"].items() if k != "ResponseMetadata"}

    def test_invoke_rest_api_failure(self, mock_conn, example_responses):
        error = ClientError(error_response=example_responses["failure"], operation_name="invoke_rest_api")
        mock_conn.invoke_rest_api = mock.MagicMock(side_effect=error)
        mock_error_log = mock.MagicMock()
        self.hook.log.error = mock_error_log

        with pytest.raises(ClientError) as caught_error:
            self.hook.invoke_rest_api(env_name=ENV_NAME, path=PATH, method=METHOD)

        assert caught_error.value == error
        mock_conn.create_web_login_token.assert_not_called()
        expected_log = {k: v for k, v in example_responses["failure"].items() if k != "ResponseMetadata"}
        mock_error_log.assert_called_once_with(expected_log)

    @pytest.mark.parametrize("generate_local_token", [pytest.param(True), pytest.param(False)])
    @mock.patch("airflow.providers.amazon.aws.hooks.mwaa.requests.Session")
    def test_invoke_rest_api_local_token_parameter(
        self, mock_create_session, generate_local_token, mock_conn
    ):
        self.hook.invoke_rest_api(
            env_name=ENV_NAME, path=PATH, method=METHOD, generate_local_token=generate_local_token
        )
        if generate_local_token:
            mock_conn.invoke_rest_api.assert_not_called()
            mock_conn.create_web_login_token.assert_called_once()
            mock_create_session.assert_called_once()
            mock_create_session.return_value.request.assert_called_once()
        else:
            mock_conn.invoke_rest_api.assert_called_once()

    @mock.patch.object(MwaaHook, "_get_session_conn")
    def test_invoke_rest_api_fallback_success_when_iam_fails_with_airflow2(
        self, mock_get_session_conn, mock_conn, example_responses
    ):
        boto_invoke_error = ClientError(
            error_response=example_responses["missingIamRole"], operation_name="invoke_rest_api"
        )
        mock_conn.invoke_rest_api = mock.MagicMock(side_effect=boto_invoke_error)

        kwargs_to_assert = {
            "method": METHOD,
            "url": f"https://{HOSTNAME}/api/v1{PATH}",
            "params": QUERY_PARAMS,
            "headers": {},
            "json": BODY,
            "timeout": 10,
        }

        mock_response = mock.MagicMock()
        mock_response.status_code = example_responses["success"]["RestApiStatusCode"]
        mock_response.json.return_value = example_responses["success"]["RestApiResponse"]
        mock_session = mock.MagicMock()
        mock_session.request.return_value = mock_response

        mock_get_session_conn.return_value = (mock_session, HOSTNAME, None)

        retval = self.hook.invoke_rest_api(
            env_name=ENV_NAME, path=PATH, method=METHOD, body=BODY, query_params=QUERY_PARAMS
        )

        mock_session.request.assert_called_once_with(**kwargs_to_assert)
        mock_response.raise_for_status.assert_called_once()
        assert retval == {k: v for k, v in example_responses["success"].items() if k != "ResponseMetadata"}

    @mock.patch.object(MwaaHook, "_get_session_conn")
    def test_invoke_rest_api_fallback_success_when_iam_fails_with_airflow3(
        self, mock_get_session_conn, mock_conn, example_responses
    ):
        boto_invoke_error = ClientError(
            error_response=example_responses["missingIamRole"], operation_name="invoke_rest_api"
        )
        mock_conn.invoke_rest_api = mock.MagicMock(side_effect=boto_invoke_error)

        kwargs_to_assert = {
            "method": METHOD,
            "url": f"https://{HOSTNAME}/api/v2{PATH}",
            "params": QUERY_PARAMS,
            "headers": {
                "Authorization": "Bearer token",
                "Content-Type": "application/json",
            },
            "json": BODY,
            "timeout": 10,
        }

        mock_response = mock.MagicMock()
        mock_response.status_code = example_responses["success"]["RestApiStatusCode"]
        mock_response.json.return_value = example_responses["success"]["RestApiResponse"]
        mock_session = mock.MagicMock()
        mock_session.request.return_value = mock_response
        mock_login_response = mock.MagicMock()
        mock_login_response.cookies = {"_token": "token"}

        mock_get_session_conn.return_value = (mock_session, HOSTNAME, mock_login_response)

        retval = self.hook.invoke_rest_api(
            env_name=ENV_NAME,
            path=PATH,
            method=METHOD,
            body=BODY,
            query_params=QUERY_PARAMS,
            airflow_version=3,
        )

        mock_session.request.assert_called_once_with(**kwargs_to_assert)
        mock_response.raise_for_status.assert_called_once()
        assert retval == {k: v for k, v in example_responses["success"].items() if k != "ResponseMetadata"}

    @mock.patch.object(MwaaHook, "_get_session_conn")
    def test_invoke_rest_api_using_local_session_token_failure(
        self, mock_get_session_conn, example_responses
    ):
        mock_response = mock.MagicMock()
        mock_response.json.return_value = example_responses["failure"]["RestApiResponse"]
        error = requests.HTTPError(response=mock_response)
        mock_response.raise_for_status.side_effect = error

        mock_session = mock.MagicMock()
        mock_session.request.return_value = mock_response

        mock_get_session_conn.return_value = (mock_session, HOSTNAME, None)

        mock_error_log = mock.MagicMock()
        self.hook.log.error = mock_error_log

        with pytest.raises(requests.HTTPError) as caught_error:
            self.hook.invoke_rest_api(env_name=ENV_NAME, path=PATH, method=METHOD, generate_local_token=True)

        assert caught_error.value == error
        mock_error_log.assert_called_once_with(example_responses["failure"]["RestApiResponse"])

    @mock.patch("airflow.providers.amazon.aws.hooks.mwaa.requests.Session")
    def test_get_session_conn_airflow2(self, mock_create_session, mock_conn):
        token = "token"
        mock_conn.create_web_login_token.return_value = {"WebServerHostname": HOSTNAME, "WebToken": token}
        login_url = f"https://{HOSTNAME}/aws_mwaa/login"
        login_payload = {"token": token}

        mock_session = mock.MagicMock()
        mock_login_response = mock.MagicMock()
        mock_session.post.return_value = mock_login_response
        mock_create_session.return_value = mock_session

        retval = self.hook._get_session_conn(env_name=ENV_NAME, airflow_version=2)

        mock_conn.create_web_login_token.assert_called_once_with(Name=ENV_NAME)
        mock_create_session.assert_called_once_with()
        mock_session.post.assert_called_once_with(login_url, data=login_payload, timeout=10)
        mock_session.post.return_value.raise_for_status.assert_called_once()

        assert retval == (mock_session, HOSTNAME, mock_login_response)

    @mock.patch("airflow.providers.amazon.aws.hooks.mwaa.requests.Session")
    def test_get_session_conn_airflow3(self, mock_create_session, mock_conn):
        token = "token"
        mock_conn.create_web_login_token.return_value = {"WebServerHostname": HOSTNAME, "WebToken": token}
        login_url = f"https://{HOSTNAME}/pluginsv2/aws_mwaa/login"
        login_payload = {"token": token}

        mock_session = mock.MagicMock()
        mock_login_response = mock.MagicMock()
        mock_session.post.return_value = mock_login_response
        mock_create_session.return_value = mock_session

        retval = self.hook._get_session_conn(env_name=ENV_NAME, airflow_version=3)

        mock_conn.create_web_login_token.assert_called_once_with(Name=ENV_NAME)
        mock_create_session.assert_called_once_with()
        mock_session.post.assert_called_once_with(login_url, data=login_payload, timeout=10)
        mock_session.post.return_value.raise_for_status.assert_called_once()

        assert retval == (mock_session, HOSTNAME, mock_login_response)

    @pytest.fixture
    def example_responses(self):
        """Fixture for test responses to avoid mutation between tests."""
        return {
            "success": {
                "ResponseMetadata": {
                    "RequestId": "some ID",
                    "HTTPStatusCode": 200,
                    "HTTPHeaders": {"header1": "value1"},
                    "RetryAttempts": 0,
                },
                "RestApiStatusCode": 200,
                "RestApiResponse": {
                    "conf": {},
                    "dag_id": "hello_world",
                    "dag_run_id": "manual__2025-02-08T00:33:09.457198+00:00",
                    "data_interval_end": "2025-02-08T00:33:09.457198+00:00",
                    "data_interval_start": "2025-02-08T00:33:09.457198+00:00",
                    "execution_date": "2025-02-08T00:33:09.457198+00:00",
                    "logical_date": "2025-02-08T00:33:09.457198+00:00",
                    "run_type": "manual",
                    "state": "queued",
                },
            },
            "failure": {
                "Error": {"Message": "", "Code": "RestApiClientException"},
                "ResponseMetadata": {
                    "RequestId": "some ID",
                    "HTTPStatusCode": 400,
                    "HTTPHeaders": {"header1": "value1"},
                    "RetryAttempts": 0,
                },
                "RestApiStatusCode": 404,
                "RestApiResponse": {
                    "detail": "DAG with dag_id: 'hello_world1' not found",
                    "status": 404,
                    "title": "DAG not found",
                    "type": "https://airflow.apache.org/docs/apache-airflow/2.10.3/stable-rest-api-ref.html#section/Errors/NotFound",
                },
            },
            "missingIamRole": {
                "Error": {"Message": "No Airflow role granted in IAM.", "Code": "AccessDeniedException"},
                "ResponseMetadata": {
                    "RequestId": "some ID",
                    "HTTPStatusCode": 403,
                    "HTTPHeaders": {"header1": "value1"},
                    "RetryAttempts": 0,
                },
            },
        }



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
from datetime import datetime

import boto3

from airflow.providers.amazon.aws.operators.batch import BatchCreateComputeEnvironmentOperator, BatchOperator
from airflow.providers.amazon.aws.operators.ecs import EcsDeregisterTaskDefinitionOperator
from airflow.providers.amazon.aws.sensors.batch import (
    BatchComputeEnvironmentSensor,
    BatchJobQueueSensor,
    BatchSensor,
)

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import (
    ENV_ID_KEY,
    SystemTestContextBuilder,
    prune_logs,
    split_string,
)

log = logging.getLogger(__name__)

DAG_ID = "example_batch"

# Externally fetched variables:
ROLE_ARN_KEY = "ROLE_ARN"
SUBNETS_KEY = "SUBNETS"
SECURITY_GROUPS_KEY = "SECURITY_GROUPS"

sys_test_context_task = (
    SystemTestContextBuilder()
    .add_variable(ROLE_ARN_KEY)
    .add_variable(SUBNETS_KEY)
    .add_variable(SECURITY_GROUPS_KEY)
    .build()
)

JOB_OVERRIDES: dict = {}


@task
def create_job_definition(role_arn, job_definition_name):
    boto3.client("batch").register_job_definition(
        type="container",
        containerProperties={
            "command": [
                "sleep",
                "2",
            ],
            "executionRoleArn": role_arn,
            "image": "busybox",
            "resourceRequirements": [
                {"value": "1", "type": "VCPU"},
                {"value": "2048", "type": "MEMORY"},
            ],
            "networkConfiguration": {
                "assignPublicIp": "ENABLED",
            },
        },
        jobDefinitionName=job_definition_name,
        platformCapabilities=["FARGATE"],
    )


@task
def create_job_queue(job_compute_environment_name, job_queue_name):
    boto3.client("batch").create_job_queue(
        computeEnvironmentOrder=[
            {
                "computeEnvironment": job_compute_environment_name,
                "order": 1,
            },
        ],
        jobQueueName=job_queue_name,
        priority=1,
        state="ENABLED",
    )


# Only describe the job if a previous task failed, to help diagnose
@task(trigger_rule=TriggerRule.ONE_FAILED)
def describe_job(job_id):
    client = boto3.client("batch")
    response = client.describe_jobs(jobs=[job_id])
    log.info("Describing the job %s for debugging purposes", job_id)
    log.info(response["jobs"])


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_job_definition(job_definition_name):
    client = boto3.client("batch")

    response = client.describe_job_definitions(
        jobDefinitionName=job_definition_name,
        status="ACTIVE",
    )

    for job_definition in response["jobDefinitions"]:
        client.deregister_job_definition(
            jobDefinition=job_definition["jobDefinitionArn"],
        )


@task(trigger_rule=TriggerRule.ALL_DONE)
def disable_compute_environment(job_compute_environment_name):
    boto3.client("batch").update_compute_environment(
        computeEnvironment=job_compute_environment_name,
        state="DISABLED",
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_compute_environment(job_compute_environment_name):
    boto3.client("batch").delete_compute_environment(
        computeEnvironment=job_compute_environment_name,
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def disable_job_queue(job_queue_name):
    boto3.client("batch").update_job_queue(
        jobQueue=job_queue_name,
        state="DISABLED",
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_job_queue(job_queue_name):
    boto3.client("batch").delete_job_queue(
        jobQueue=job_queue_name,
    )


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]

    batch_job_name: str = f"{env_id}-test-job"
    batch_job_definition_name: str = f"{env_id}-test-job-definition"
    batch_job_compute_environment_name: str = f"{env_id}-test-job-compute-environment"
    batch_job_queue_name: str = f"{env_id}-test-job-queue"

    security_groups = split_string(test_context[SECURITY_GROUPS_KEY])
    subnets = split_string(test_context[SUBNETS_KEY])

    # [START howto_operator_batch_create_compute_environment]
    create_compute_environment = BatchCreateComputeEnvironmentOperator(
        task_id="create_compute_environment",
        compute_environment_name=batch_job_compute_environment_name,
        environment_type="MANAGED",
        state="ENABLED",
        compute_resources={
            "type": "FARGATE",
            "maxvCpus": 10,
            "securityGroupIds": security_groups,
            "subnets": subnets,
        },
    )
    # [END howto_operator_batch_create_compute_environment]

    # [START howto_sensor_batch_compute_environment]
    wait_for_compute_environment_valid = BatchComputeEnvironmentSensor(
        task_id="wait_for_compute_environment_valid",
        compute_environment=batch_job_compute_environment_name,
    )
    # [END howto_sensor_batch_compute_environment]
    wait_for_compute_environment_valid.poke_interval = 1

    # [START howto_sensor_batch_job_queue]
    wait_for_job_queue_valid = BatchJobQueueSensor(
        task_id="wait_for_job_queue_valid",
        job_queue=batch_job_queue_name,
    )
    # [END howto_sensor_batch_job_queue]
    wait_for_job_queue_valid.poke_interval = 1

    # [START howto_operator_batch]
    submit_batch_job = BatchOperator(
        task_id="submit_batch_job",
        job_name=batch_job_name,
        job_queue=batch_job_queue_name,
        job_definition=batch_job_definition_name,
        container_overrides=JOB_OVERRIDES,
    )
    # [END howto_operator_batch]

    # BatchOperator waits by default, setting as False to test the Sensor below.
    submit_batch_job.wait_for_completion = False

    # [START howto_sensor_batch]
    wait_for_batch_job = BatchSensor(
        task_id="wait_for_batch_job",
        job_id=submit_batch_job.output,
    )
    # [END howto_sensor_batch]
    wait_for_batch_job.poke_interval = 10

    wait_for_compute_environment_disabled = BatchComputeEnvironmentSensor(
        task_id="wait_for_compute_environment_disabled",
        compute_environment=batch_job_compute_environment_name,
        poke_interval=1,
    )

    wait_for_job_queue_modified = BatchJobQueueSensor(
        task_id="wait_for_job_queue_modified",
        job_queue=batch_job_queue_name,
        poke_interval=1,
    )

    wait_for_job_queue_deleted = BatchJobQueueSensor(
        task_id="wait_for_job_queue_deleted",
        job_queue=batch_job_queue_name,
        treat_non_existing_as_deleted=True,
        poke_interval=10,
    )

    deregister_task_definition = EcsDeregisterTaskDefinitionOperator(
        task_id="deregister_task",
        task_definition=f"{batch_job_definition_name}:1",
        trigger_rule=TriggerRule.ALL_DONE,
    )

    log_cleanup = prune_logs(
        [
            # Format: ('log group name', 'log stream prefix')
            ("/aws/batch/job", env_id)
        ],
    )

    chain(
        # TEST SETUP
        test_context,
        security_groups,
        subnets,
        create_job_definition(test_context[ROLE_ARN_KEY], batch_job_definition_name),
        # TEST BODY
        create_compute_environment,
        wait_for_compute_environment_valid,
        # ``create_job_queue`` is part of test setup but need the compute-environment to be created before
        create_job_queue(batch_job_compute_environment_name, batch_job_queue_name),
        wait_for_job_queue_valid,
        submit_batch_job,
        wait_for_batch_job,
        # TEST TEARDOWN
        describe_job(submit_batch_job.output),
        disable_job_queue(batch_job_queue_name),
        wait_for_job_queue_modified,
        delete_job_queue(batch_job_queue_name),
        wait_for_job_queue_deleted,
        disable_compute_environment(batch_job_compute_environment_name),
        wait_for_compute_environment_disabled,
        delete_compute_environment(batch_job_compute_environment_name),
        delete_job_definition(batch_job_definition_name),
        deregister_task_definition,
        log_cleanup,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
from typing import TYPE_CHECKING, cast
from unittest.mock import patch

import pytest
import responses
from botocore.awsrequest import AWSRequest

if TYPE_CHECKING:
    import s3fs

pytest.importorskip("s3fs")

TEST_CONN = "aws_test_conn"
TEST_SIGNER_URL = "https://nowhere.to.be.found"
TEST_SIGNER_TOKEN = "TOKEN1234"
TEST_SIGNER_RESP_URL = "https://where.are.you"
TEST_HEADER_KEY = "x-airflow"
TEST_HEADER_VALUE = "payload"
TEST_REQ_URI = "s3://bucket/key"


@pytest.fixture(scope="module", autouse=True)
def _setup_connections():
    with pytest.MonkeyPatch.context() as mp_ctx:
        mp_ctx.setenv(f"AIRFLOW_CONN_{TEST_CONN}".upper(), "aws://")
        yield


class TestFilesystem:
    def test_get_s3fs(self):
        from airflow.providers.amazon.aws.fs.s3 import get_fs

        fs = get_fs(conn_id=TEST_CONN, storage_options={"key": "value"})
        fs = cast("s3fs.S3FileSystem", fs)

        assert "s3" in fs.protocol
        assert fs.config_kwargs["key"] == "value"

    @patch("s3fs.S3FileSystem", autospec=True)
    def test_get_s3fs_anonymous(self, s3fs, monkeypatch):
        from airflow.providers.amazon.aws.fs.s3 import get_fs

        # remove all AWS_* env vars
        for env_name in os.environ:
            if env_name.startswith("AWS"):
                monkeypatch.delenv(env_name, raising=False)

        get_fs(conn_id=None, storage_options=None)

        assert s3fs.call_args.kwargs["anon"] is True

    @responses.activate
    def test_signer(self):
        from airflow.providers.amazon.aws.fs.s3 import s3v4_rest_signer

        req = AWSRequest(
            method="GET",
            url=TEST_REQ_URI,
            headers={"x": "y"},
        )
        req.context = {"client_region": "antarctica"}

        responses.add(
            responses.POST,
            f"{TEST_SIGNER_URL}/v1/aws/s3/sign",
            json={
                "uri": TEST_SIGNER_RESP_URL,
                "headers": {
                    TEST_HEADER_KEY: [TEST_HEADER_VALUE],
                },
            },
        )

        req = s3v4_rest_signer(
            {
                "uri": TEST_SIGNER_URL,
                "token": TEST_SIGNER_TOKEN,
            },
            req,
        )

        assert req.url == TEST_SIGNER_RESP_URL
        assert req.headers[TEST_HEADER_KEY] == TEST_HEADER_VALUE



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
AWS ECS Executor.

Each Airflow task gets delegated out to an Amazon ECS Task.
"""

from __future__ import annotations

import time
from collections import defaultdict, deque
from collections.abc import Sequence
from copy import deepcopy
from typing import TYPE_CHECKING

from botocore.exceptions import ClientError, NoCredentialsError

from airflow.executors.base_executor import BaseExecutor
from airflow.providers.amazon.aws.executors.ecs.boto_schema import BotoDescribeTasksSchema, BotoRunTaskSchema
from airflow.providers.amazon.aws.executors.ecs.utils import (
    CONFIG_DEFAULTS,
    CONFIG_GROUP_NAME,
    AllEcsConfigKeys,
    EcsExecutorException,
    EcsQueuedTask,
    EcsTaskCollection,
)
from airflow.providers.amazon.aws.executors.utils.exponential_backoff_retry import (
    calculate_next_attempt_delay,
    exponential_backoff_retry,
)
from airflow.providers.amazon.aws.hooks.ecs import EcsHook
from airflow.providers.amazon.version_compat import AIRFLOW_V_3_0_PLUS
from airflow.providers.common.compat.sdk import AirflowException, Stats, timezone
from airflow.utils.helpers import merge_dicts
from airflow.utils.state import State

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from airflow.executors import workloads
    from airflow.models.taskinstance import TaskInstance, TaskInstanceKey
    from airflow.providers.amazon.aws.executors.ecs.utils import (
        CommandType,
        ExecutorConfigType,
    )

INVALID_CREDENTIALS_EXCEPTIONS = [
    "ExpiredTokenException",
    "InvalidClientTokenId",
    "UnrecognizedClientException",
]


class AwsEcsExecutor(BaseExecutor):
    """
    Executes the provided Airflow command on an ECS instance.

    The Airflow Scheduler creates a shell command, and passes it to the executor. This ECS Executor
    runs said Airflow command on a remote Amazon ECS Cluster with a task-definition configured to
    launch the same containers as the Scheduler. It then periodically checks in with the launched
    tasks (via task ARNs) to determine the status.

    This allows individual tasks to specify CPU, memory, GPU, env variables, etc. When initializing a task,
    there's an option for "executor config" which should be a dictionary with keys that match the
    ``ContainerOverride`` definition per AWS documentation (see link below).

    Prerequisite: proper configuration of Boto3 library
    .. seealso:: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html for
    authentication and access-key management. You can store an environmental variable, setup aws config from
    console, or use IAM roles.

    .. seealso:: https://docs.aws.amazon.com/AmazonECS/latest/APIReference/API_ContainerOverride.html for an
     Airflow TaskInstance's executor_config.
    """

    supports_multi_team: bool = True

    # AWS limits the maximum number of ARNs in the describe_tasks function.
    DESCRIBE_TASKS_BATCH_SIZE = 99

    if TYPE_CHECKING and AIRFLOW_V_3_0_PLUS:
        # In the v3 path, we store workloads, not commands as strings.
        # TODO: TaskSDK: move this type change into BaseExecutor
        queued_tasks: dict[TaskInstanceKey, workloads.All]  # type: ignore[assignment]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.active_workers: EcsTaskCollection = EcsTaskCollection()
        self.pending_tasks: deque = deque()

        # Check if self has the ExecutorConf set on the self.conf attribute, and if not, set it to the global
        # configuration object. This allows the changes to be backwards compatible with older versions of
        # Airflow.
        # Can be removed when minimum supported provider version is equal to the version of core airflow
        # which introduces multi-team configuration.
        if not hasattr(self, "conf"):
            from airflow.providers.common.compat.sdk import conf

            self.conf = conf

        self.cluster = self.conf.get(CONFIG_GROUP_NAME, AllEcsConfigKeys.CLUSTER)
        self.container_name = self.conf.get(CONFIG_GROUP_NAME, AllEcsConfigKeys.CONTAINER_NAME)
        self.attempts_since_last_successful_connection = 0

        self.load_ecs_connection(check_connection=False)
        self.IS_BOTO_CONNECTION_HEALTHY = False

        self.run_task_kwargs = self._load_run_kwargs()

        # Maximum number of retries to run an ECS task.
        self.max_run_task_attempts = self.conf.get(
            CONFIG_GROUP_NAME,
            AllEcsConfigKeys.MAX_RUN_TASK_ATTEMPTS,
            fallback=CONFIG_DEFAULTS[AllEcsConfigKeys.MAX_RUN_TASK_ATTEMPTS],
        )

    def queue_workload(self, workload: workloads.All, session: Session | None) -> None:
        from airflow.executors import workloads

        if not isinstance(workload, workloads.ExecuteTask):
            raise RuntimeError(f"{type(self)} cannot handle workloads of type {type(workload)}")
        ti = workload.ti
        self.queued_tasks[ti.key] = workload

    def _process_workloads(self, workloads: Sequence[workloads.All]) -> None:
        from airflow.executors.workloads import ExecuteTask

        # Airflow V3 version
        for w in workloads:
            if not isinstance(w, ExecuteTask):
                raise RuntimeError(f"{type(self)} cannot handle workloads of type {type(w)}")

            command = [w]
            key = w.ti.key
            queue = w.ti.queue
            executor_config = w.ti.executor_config or {}

            del self.queued_tasks[key]
            self.execute_async(key=key, command=command, queue=queue, executor_config=executor_config)  # type: ignore[arg-type]
            self.running.add(key)

    def start(self):
        """Call this when the Executor is run for the first time by the scheduler."""
        check_health = self.conf.getboolean(
            CONFIG_GROUP_NAME, AllEcsConfigKeys.CHECK_HEALTH_ON_STARTUP, fallback=False
        )

        if not check_health:
            return

        self.log.info("Starting ECS Executor and determining health...")
        try:
            self.check_health()
        except AirflowException:
            self.log.error("Stopping the Airflow Scheduler from starting until the issue is resolved.")
            raise

    def check_health(self):
        """
        Make a test API call to check the health of the ECS Executor.

        Deliberately use an invalid task ID, some potential outcomes in order:
          1. `AccessDeniedException` is raised if there are insufficient permissions.
          2. `ClusterNotFoundException` is raised if permissions exist but the cluster does not.
          3. The API responds with a failure message if the cluster is found and there
             are permissions, but the cluster itself has issues.
          4. `InvalidParameterException` is raised if the permissions and cluster exist but the task does not.

        The last one is considered a success state for the purposes of this check.
        """
        success_status = "succeeded."
        status = success_status

        try:
            invalid_task_id = "a" * 32
            self.ecs.stop_task(cluster=self.cluster, task=invalid_task_id)

            # If it got this far, something is wrong.  stop_task() called with an
            # invalid taskID should have thrown a ClientError.  All known reasons are
            # covered in the ``except`` block below, and this should never be reached.
            status = "failed for an unknown reason. "
        except ClientError as ex:
            error_code = ex.response["Error"]["Code"]
            error_message = ex.response["Error"]["Message"]

            if ("InvalidParameterException" in error_code) and ("task was not found" in error_message):
                # This failure is expected, and means we're healthy
                pass
            else:
                # Catch all for unexpected failures
                status = f"failed because: {error_message}. "
        except Exception as e:
            # Any non-ClientError exceptions. This can include Botocore exceptions for example
            status = f"failed because: {e}. "
        finally:
            msg_prefix = "ECS Executor health check has %s"
            if status == success_status:
                self.IS_BOTO_CONNECTION_HEALTHY = True
                self.log.info(msg_prefix, status)
            else:
                msg_error_suffix = (
                    "The ECS executor will not be able to run Airflow tasks until the issue is addressed."
                )
                raise AirflowException(msg_prefix % status + msg_error_suffix)

    def load_ecs_connection(self, check_connection: bool = True):
        self.log.info("Loading Connection information")
        aws_conn_id = self.conf.get(
            CONFIG_GROUP_NAME,
            AllEcsConfigKeys.AWS_CONN_ID,
            fallback=CONFIG_DEFAULTS[AllEcsConfigKeys.AWS_CONN_ID],
        )
        region_name = self.conf.get(CONFIG_GROUP_NAME, AllEcsConfigKeys.REGION_NAME, fallback=None)
        self.ecs = EcsHook(aws_conn_id=aws_conn_id, region_name=region_name).conn
        self.attempts_since_last_successful_connection += 1
        self.last_connection_reload = timezone.utcnow()

        if check_connection:
            self.check_health()
            self.attempts_since_last_successful_connection = 0

    def sync(self):
        if not self.IS_BOTO_CONNECTION_HEALTHY:
            exponential_backoff_retry(
                self.last_connection_reload,
                self.attempts_since_last_successful_connection,
                self.load_ecs_connection,
            )
            if not self.IS_BOTO_CONNECTION_HEALTHY:
                return
        try:
            self.sync_running_tasks()
            self.attempt_task_runs()
        except (ClientError, NoCredentialsError) as error:
            error_code = error.response["Error"]["Code"]
            if error_code in INVALID_CREDENTIALS_EXCEPTIONS:
                self.IS_BOTO_CONNECTION_HEALTHY = False
                self.log.warning(
                    "AWS credentials are either missing or expired: %s.\nRetrying connection", error
                )

        except Exception:
            # We catch any and all exceptions because otherwise they would bubble
            # up and kill the scheduler process
            self.log.exception("Failed to sync %s", self.__class__.__name__)

    def sync_running_tasks(self):
        """Check and update state on all running tasks."""
        all_task_arns = self.active_workers.get_all_arns()
        if not all_task_arns:
            self.log.debug("No active Airflow tasks, skipping sync.")
            return

        describe_tasks_response = self.__describe_tasks(all_task_arns)

        self.log.debug("Active Workers: %s", describe_tasks_response)

        if describe_tasks_response["failures"]:
            for failure in describe_tasks_response["failures"]:
                self.__handle_failed_task(failure["arn"], failure["reason"])

        updated_tasks = describe_tasks_response["tasks"]
        for task in updated_tasks:
            self.__update_running_task(task)

    def __update_running_task(self, task):
        self.active_workers.update_task(task)
        # Get state of current task.
        task_state = task.get_task_state()
        task_key = self.active_workers.arn_to_key[task.task_arn]

        # Mark finished tasks as either a success/failure.
        if task_state == State.FAILED or task_state == State.REMOVED:
            self.__log_container_failures(task_arn=task.task_arn)
            self.__handle_failed_task(task.task_arn, task.stopped_reason)
        elif task_state == State.SUCCESS:
            self.log.debug(
                "Airflow task %s marked as %s after running on ECS Task (arn) %s",
                task_key,
                task_state,
                task.task_arn,
            )
            self.success(task_key)
            self.active_workers.pop_by_key(task_key)

    def __describe_tasks(self, task_arns):
        all_task_descriptions = {"tasks": [], "failures": []}
        for i in range(0, len(task_arns), self.DESCRIBE_TASKS_BATCH_SIZE):
            batched_task_arns = task_arns[i : i + self.DESCRIBE_TASKS_BATCH_SIZE]
            if not batched_task_arns:
                continue
            boto_describe_tasks = self.ecs.describe_tasks(tasks=batched_task_arns, cluster=self.cluster)
            describe_tasks_response = BotoDescribeTasksSchema().load(boto_describe_tasks)

            all_task_descriptions["tasks"].extend(describe_tasks_response["tasks"])
            all_task_descriptions["failures"].extend(describe_tasks_response["failures"])
        return all_task_descriptions

    def __log_container_failures(self, task_arn: str):
        """Check if the task failed due to issues with the containers."""
        containers = self.active_workers.task_by_arn(task_arn).containers
        has_exit_codes = all(["exit_code" in x for x in containers])
        if not has_exit_codes:
            return ""
        reasons = [
            f"{container['container_arn']} - {container['reason']}"
            for container in containers
            if "reason" in container
        ]
        if reasons:
            self.log.warning(
                "The ECS task failed due to the following containers failing:\n%s", "\n".join(reasons)
            )

    def __handle_failed_task(self, task_arn: str, reason: str):
        """
        If an API failure occurs, the task is rescheduled.

        This function will determine whether the task has been attempted the appropriate number
        of times, and determine whether the task should be marked failed or not. The task will
        be removed active_workers, and marked as FAILED, or set into pending_tasks depending on
        how many times it has been retried.
        """
        task_key = self.active_workers.arn_to_key[task_arn]
        task_info = self.active_workers.info_by_key(task_key)
        task_cmd = task_info.cmd
        queue = task_info.queue
        exec_info = task_info.config
        failure_count = self.active_workers.failure_count_by_key(task_key)
        if int(failure_count) < int(self.max_run_task_attempts):
            self.log.warning(
                "Airflow task %s failed due to %s. Failure %s out of %s occurred on %s. Rescheduling.",
                task_key,
                reason,
                failure_count,
                self.max_run_task_attempts,
                task_arn,
            )
            self.pending_tasks.append(
                EcsQueuedTask(
                    task_key,
                    task_cmd,
                    queue,
                    exec_info,
                    failure_count + 1,
                    timezone.utcnow() + calculate_next_attempt_delay(failure_count),
                )
            )
        else:
            self.log.error(
                "Airflow task %s has failed a maximum of %s times. Marking as failed",
                task_key,
                failure_count,
            )
            self.fail(task_key)
        self.active_workers.pop_by_key(task_key)

    def attempt_task_runs(self):
        """
        Take tasks from the pending_tasks queue, and attempts to find an instance to run it on.

        If the launch type is EC2, this will attempt to place tasks on empty EC2 instances.  If
            there are no EC2 instances available, no task is placed and this function will be
            called again in the next heart-beat.

        If the launch type is FARGATE, this will run the tasks on new AWS Fargate instances.
        """
        queue_len = len(self.pending_tasks)
        failure_reasons = defaultdict(int)
        for _ in range(queue_len):
            ecs_task = self.pending_tasks.popleft()
            task_key = ecs_task.key
            cmd = ecs_task.command
            queue = ecs_task.queue
            exec_config = ecs_task.executor_config
            attempt_number = ecs_task.attempt_number
            failure_reasons = []
            if timezone.utcnow() < ecs_task.next_attempt_time:
                self.pending_tasks.append(ecs_task)
                continue
            try:
                run_task_response = self._run_task(task_key, cmd, queue, exec_config)
            except NoCredentialsError:
                self.pending_tasks.append(ecs_task)
                raise
            except ClientError as e:
                error_code = e.response["Error"]["Code"]
                if error_code in INVALID_CREDENTIALS_EXCEPTIONS:
                    self.pending_tasks.append(ecs_task)
                    raise
                failure_reasons.append(str(e))
            except Exception as e:
                # Failed to even get a response back from the Boto3 API or something else went
                # wrong.  For any possible failure we want to add the exception reasons to the
                # failure list so that it is logged to the user and most importantly the task is
                # added back to the pending list to be retried later.
                failure_reasons.append(str(e))
            else:
                # We got a response back, check if there were failures. If so, add them to the
                # failures list so that it is logged to the user and most importantly the task
                # is added back to the pending list to be retried later.
                if run_task_response["failures"]:
                    failure_reasons.extend([f["reason"] for f in run_task_response["failures"]])

            if failure_reasons:
                # Make sure the number of attempts does not exceed max_run_task_attempts
                if int(attempt_number) < int(self.max_run_task_attempts):
                    ecs_task.attempt_number += 1
                    ecs_task.next_attempt_time = timezone.utcnow() + calculate_next_attempt_delay(
                        attempt_number
                    )
                    self.pending_tasks.append(ecs_task)
                else:
                    reasons_str = ", ".join(failure_reasons)
                    self.log.error(
                        "ECS task %s has failed a maximum of %s times. Marking as failed. Reasons: %s",
                        task_key,
                        attempt_number,
                        reasons_str,
                    )
                    self.log_task_event(
                        event="ecs task submit failure",
                        ti_key=task_key,
                        extra=(
                            f"Task could not be queued after {attempt_number} attempts. "
                            f"Marking as failed. Reasons: {reasons_str}"
                        ),
                    )
                    self.fail(task_key)
            elif not run_task_response["tasks"]:
                self.log.error("ECS RunTask Response: %s", run_task_response)
                self.log_task_event(
                    event="ecs task submit failure",
                    extra=f"ECS RunTask Response: {run_task_response}",
                    ti_key=task_key,
                )
                raise EcsExecutorException(
                    "No failures and no ECS tasks provided in response. This should never happen."
                )
            else:
                task = run_task_response["tasks"][0]
                self.active_workers.add_task(task, task_key, queue, cmd, exec_config, attempt_number)
                self.running_state(task_key, task.task_arn)

    def _run_task(
        self, task_id: TaskInstanceKey, cmd: CommandType, queue: str, exec_config: ExecutorConfigType
    ):
        """
        Run a queued-up Airflow task.

        Not to be confused with execute_async() which inserts tasks into the queue.
        The command and executor config will be placed in the container-override
        section of the JSON request before calling Boto3's "run_task" function.
        """
        run_task_kwargs = self._run_task_kwargs(task_id, cmd, queue, exec_config)
        boto_run_task = self.ecs.run_task(**run_task_kwargs)
        run_task_response = BotoRunTaskSchema().load(boto_run_task)
        return run_task_response

    def _run_task_kwargs(
        self, task_id: TaskInstanceKey, cmd: CommandType, queue: str, exec_config: ExecutorConfigType
    ) -> dict:
        """
        Update the Airflow command by modifying container overrides for task-specific kwargs.

        One last chance to modify Boto3's "run_task" kwarg params before it gets passed into the Boto3 client.
        """
        run_task_kwargs = deepcopy(self.run_task_kwargs)
        run_task_kwargs = merge_dicts(run_task_kwargs, exec_config)
        container_override = self.get_container(run_task_kwargs["overrides"]["containerOverrides"])
        container_override["command"] = cmd

        # Inject the env variable to configure logging for containerized execution environment
        if "environment" not in container_override:
            container_override["environment"] = []
        container_override["environment"].append({"name": "AIRFLOW_IS_EXECUTOR_CONTAINER", "value": "true"})

        return run_task_kwargs

    def execute_async(self, key: TaskInstanceKey, command: CommandType, queue=None, executor_config=None):
        """Save the task to be executed in the next sync by inserting the commands into a queue."""
        if executor_config and ("name" in executor_config or "command" in executor_config):
            raise ValueError('Executor Config should never override "name" or "command"')
        if len(command) == 1:
            from airflow.executors.workloads import ExecuteTask

            if isinstance(command[0], ExecuteTask):
                workload = command[0]
                ser_input = workload.model_dump_json()
                command = [
                    "python",
                    "-m",
                    "airflow.sdk.execution_time.execute_workload",
                    "--json-string",
                    ser_input,
                ]
            else:
                raise ValueError(
                    f"EcsExecutor doesn't know how to handle workload of type: {type(command[0])}"
                )

        self.pending_tasks.append(
            EcsQueuedTask(key, command, queue, executor_config or {}, 1, timezone.utcnow())
        )

    def end(self, heartbeat_interval=10):
        """Wait for all currently running tasks to end, and don't launch any tasks."""
        try:
            while True:
                self.sync()
                if not self.active_workers:
                    break
                time.sleep(heartbeat_interval)
        except Exception:
            # We catch any and all exceptions because otherwise they would bubble
            # up and kill the scheduler process.
            self.log.exception("Failed to end %s", self.__class__.__name__)

    def terminate(self):
        """Kill all ECS processes by calling Boto3's StopTask API."""
        try:
            for arn in self.active_workers.get_all_arns():
                self.ecs.stop_task(
                    cluster=self.cluster, task=arn, reason="Airflow Executor received a SIGTERM"
                )
            self.end()
        except Exception:
            # We catch any and all exceptions because otherwise they would bubble
            # up and kill the scheduler process.
            self.log.exception("Failed to terminate %s", self.__class__.__name__)

    def _load_run_kwargs(self) -> dict:
        from airflow.providers.amazon.aws.executors.ecs.ecs_executor_config import build_task_kwargs

        ecs_executor_run_task_kwargs = build_task_kwargs(self.conf)

        try:
            self.get_container(ecs_executor_run_task_kwargs["overrides"]["containerOverrides"])["command"]
        except KeyError:
            raise KeyError(
                "Rendered JSON template does not contain key "
                '"overrides[containerOverrides][containers][x][command]"'
            )
        return ecs_executor_run_task_kwargs

    def get_container(self, container_list):
        """Search task list for core Airflow container."""
        for container in container_list:
            try:
                if container["name"] == self.container_name:
                    return container
            except KeyError:
                raise EcsExecutorException(
                    'container "name" must be provided in "containerOverrides" configuration'
                )
        raise KeyError(f"No such container found by container name: {self.container_name}")

    def try_adopt_task_instances(self, tis: Sequence[TaskInstance]) -> Sequence[TaskInstance]:
        """
        Adopt task instances which have an external_executor_id (the ECS task ARN).

        Anything that is not adopted will be cleared by the scheduler and becomes eligible for re-scheduling.
        """
        with Stats.timer("ecs_executor.adopt_task_instances.duration"):
            adopted_tis: list[TaskInstance] = []

            if task_arns := [ti.external_executor_id for ti in tis if ti.external_executor_id]:
                task_descriptions = self.__describe_tasks(task_arns).get("tasks", [])

                for task in task_descriptions:
                    ti = next(ti for ti in tis if ti.external_executor_id == task.task_arn)
                    self.active_workers.add_task(
                        task,
                        ti.key,
                        ti.queue,
                        ti.command_as_list(),
                        ti.executor_config,
                        ti.try_number,
                    )
                    adopted_tis.append(ti)

            if adopted_tis:
                tasks = [f"{task} in state {task.state}" for task in adopted_tis]
                task_instance_str = "\n\t".join(tasks)
                self.log.info(
                    "Adopted the following %d tasks from a dead executor:\n\t%s",
                    len(adopted_tis),
                    task_instance_str,
                )

            not_adopted_tis = [ti for ti in tis if ti not in adopted_tis]
            return not_adopted_tis



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
import time
from datetime import datetime, timedelta, timezone
from urllib.parse import urlparse

import boto3

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

from system.amazon.aws.utils import SystemTestContextBuilder

sys_test_context_task = SystemTestContextBuilder().build()

DAG_ID = "test_lambda_executor_dlq"


@task
def verify_test_setup():
    """Verify Lambda executor DLQ configuration is available"""
    dlq_url = os.environ.get("AIRFLOW__AWS_LAMBDA_EXECUTOR__DEAD_LETTER_QUEUE_URL")
    print(f"Lambda executor DLQ URL: {dlq_url}")

    parsed_url = urlparse(dlq_url)
    dlq_queue_name = parsed_url.path.split("/")[-1]
    return dlq_queue_name


# Configuration to search for this poison pill has been done within the Lambda app.py handler code
@task(executor_config={"poison_pill": True})
def cause_lambda_failure():
    return "This task is designed to fail at Lambda service level, injected within the function handler"


@task(trigger_rule=TriggerRule.ALL_DONE)
def verify_dlq_activity(dlq_queue_name: str):
    """Verify DLQ processing occurred by checking CloudWatch metrics"""

    cloudwatch = boto3.client("cloudwatch")
    # Try for up to 10 attempts (5 minutes total)
    for attempt in range(10):
        end_time = datetime.now(timezone.utc)
        start_time = end_time - timedelta(minutes=5)
        received_response = cloudwatch.get_metric_statistics(
            Namespace="AWS/SQS",
            MetricName="NumberOfMessagesDeleted",
            Dimensions=[{"Name": "QueueName", "Value": dlq_queue_name}],
            StartTime=start_time,
            EndTime=end_time,
            Period=300,
            Statistics=["Sum"],
        )

        total_deleted = sum(point["Sum"] for point in received_response["Datapoints"])
        if total_deleted > 0:
            print(f"Messages deleted from DLQ: {total_deleted}")
            return "DLQ Processing Confirmed"
        print("No messages detected in DLQ yet")
        if attempt < 9:
            time.sleep(30)
    raise AssertionError("FAIL: No DLQ activity detected after 5 minutes of polling")


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    tags=["test"],
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    dlq_name = verify_test_setup()
    failed_task = cause_lambda_failure()
    verify_task = verify_dlq_activity(dlq_name)

    chain(
        test_context,
        dlq_name,
        failed_task,
        verify_task,
    )
    from tests_common.test_utils.watcher import watcher

    verify_task >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import boto3
import pytest
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.ses import SesHook

boto3.setup_default_session()


TEST_TO_ADDRESSES = [
    pytest.param("to@domain.com", id="to=single_string"),
    pytest.param("to1@domain.com,to2@domain.com", id="to=comma_string"),
    pytest.param(["to1@domain.com", "to2@domain.com"], id="to=list"),
]

TEST_CC_ADDRESSES = [
    pytest.param("cc@domain.com", id="cc=single_string"),
    pytest.param("cc1@domain.com,cc2@domain.com", id="cc=comma_string"),
    pytest.param(["cc1@domain.com", "cc2@domain.com"], id="cc=list"),
]

TEST_BCC_ADDRESSES = [
    pytest.param("bcc@domain.com", id="bcc=single_string"),
    pytest.param("bcc1@domain.com,bcc2@domain.com", id="bcc=comma_string"),
    pytest.param(["bcc1@domain.com", "bcc2@domain.com"], id="bcc=list"),
]

TEST_FROM_ADDRESS = "test_from@domain.com"
TEST_SUBJECT = "subject"
TEST_HTML_CONTENT = "<html>Test</html>"
TEST_REPLY_TO = "reply_to@domain.com"
TEST_RETURN_PATH = "return_path@domain.com"


@mock_aws
def _verify_address(address: str) -> None:
    """
    Amazon only allows emails from verified addresses. If the address is not verified,
    this test will raise `botocore.errorfactory.MessageRejected`.
    """
    SesHook().get_conn().verify_email_identity(EmailAddress=address)


@mock_aws
class TestSesHook:
    def test_get_conn(self):
        hook = SesHook(aws_conn_id="aws_default")
        assert hook.get_conn() is not None

    @pytest.mark.parametrize("to", TEST_TO_ADDRESSES)
    @pytest.mark.parametrize("cc", TEST_CC_ADDRESSES)
    @pytest.mark.parametrize("bcc", TEST_BCC_ADDRESSES)
    def test_send_email(self, to, cc, bcc):
        _verify_address(TEST_FROM_ADDRESS)
        hook = SesHook()

        response = hook.send_email(
            mail_from=TEST_FROM_ADDRESS,
            to=to,
            subject=TEST_SUBJECT,
            html_content=TEST_HTML_CONTENT,
            cc=cc,
            bcc=bcc,
            reply_to=TEST_REPLY_TO,
            return_path=TEST_RETURN_PATH,
        )

        assert response is not None
        assert isinstance(response, dict)
        assert "MessageId" in response


@pytest.mark.asyncio
class TestAsyncSesHook:
    """The mock_aws decorator uses `moto` which does not currently support async SES so we mock it manually."""

    @pytest.fixture
    def mock_async_client(self):
        return mock.AsyncMock()

    @pytest.fixture
    def mock_get_async_conn(self, mock_async_client):
        with mock.patch.object(SesHook, "get_async_conn") as mocked_conn:
            mocked_conn.return_value.__aenter__.return_value = mock_async_client
            yield mocked_conn

    async def test_get_async_conn(self, mock_get_async_conn, mock_async_client):
        hook = SesHook()
        async with await hook.get_async_conn() as async_conn:
            assert async_conn is mock_async_client

    @pytest.mark.parametrize("to", TEST_TO_ADDRESSES)
    @pytest.mark.parametrize("cc", TEST_CC_ADDRESSES)
    @pytest.mark.parametrize("bcc", TEST_BCC_ADDRESSES)
    async def test_asend_email(self, mock_get_async_conn, mock_async_client, to, cc, bcc):
        _verify_address(TEST_FROM_ADDRESS)
        hook = SesHook()

        mock_async_client.send_raw_email.return_value = {"MessageId": "test_message_id"}

        response = await hook.asend_email(
            mail_from=TEST_FROM_ADDRESS,
            to=to,
            subject=TEST_SUBJECT,
            html_content=TEST_HTML_CONTENT,
            cc=cc,
            bcc=bcc,
            reply_to=TEST_REPLY_TO,
            return_path=TEST_RETURN_PATH,
        )

        assert response is not None
        assert isinstance(response, dict)
        assert "MessageId" in response
        assert response["MessageId"] == "test_message_id"

        mock_async_client.send_raw_email.assert_called_once()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.providers.amazon.aws.hooks.bedrock import BedrockAgentHook, BedrockHook
from airflow.providers.amazon.aws.triggers.base import AwsBaseWaiterTrigger
from airflow.providers.amazon.version_compat import NOTSET, ArgNotSet

if TYPE_CHECKING:
    from airflow.providers.amazon.aws.hooks.base_aws import AwsGenericHook


class BedrockCustomizeModelCompletedTrigger(AwsBaseWaiterTrigger):
    """
    Trigger when a Bedrock model customization job is complete.

    :param job_name: The name of the Bedrock model customization job.
    :param waiter_delay: The amount of time in seconds to wait between attempts. (default: 120)
    :param waiter_max_attempts: The maximum number of attempts to be made. (default: 75)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        *,
        job_name: str,
        waiter_delay: int = 120,
        waiter_max_attempts: int = 75,
        aws_conn_id: str | None = None,
    ) -> None:
        super().__init__(
            serialized_fields={"job_name": job_name},
            waiter_name="model_customization_job_complete",
            waiter_args={"jobIdentifier": job_name},
            failure_message="Bedrock model customization failed.",
            status_message="Status of Bedrock model customization job is",
            status_queries=["status"],
            return_key="job_name",
            return_value=job_name,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )

    def hook(self) -> AwsGenericHook:
        return BedrockHook(aws_conn_id=self.aws_conn_id)


class BedrockKnowledgeBaseActiveTrigger(AwsBaseWaiterTrigger):
    """
    Trigger when a Bedrock Knowledge Base reaches the ACTIVE state.

    :param knowledge_base_id: The unique identifier of the knowledge base for which to get information.

    :param waiter_delay: The amount of time in seconds to wait between attempts. (default: 5)
    :param waiter_max_attempts: The maximum number of attempts to be made. (default: 24)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        *,
        knowledge_base_id: str,
        waiter_delay: int = 5,
        waiter_max_attempts: int = 24,
        aws_conn_id: str | None = None,
    ) -> None:
        super().__init__(
            serialized_fields={"knowledge_base_id": knowledge_base_id},
            waiter_name="knowledge_base_active",
            waiter_args={"knowledgeBaseId": knowledge_base_id},
            failure_message="Bedrock Knowledge Base creation failed.",
            status_message="Status of Bedrock Knowledge Base job is",
            status_queries=["status"],
            return_key="knowledge_base_id",
            return_value=knowledge_base_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )

    def hook(self) -> AwsGenericHook:
        return BedrockAgentHook(aws_conn_id=self.aws_conn_id)


class BedrockProvisionModelThroughputCompletedTrigger(AwsBaseWaiterTrigger):
    """
    Trigger when a provisioned throughput job is complete.

    :param provisioned_model_id: The ARN or name of the provisioned throughput.

    :param waiter_delay: The amount of time in seconds to wait between attempts. (default: 120)
    :param waiter_max_attempts: The maximum number of attempts to be made. (default: 75)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        *,
        provisioned_model_id: str,
        waiter_delay: int = 120,
        waiter_max_attempts: int = 75,
        aws_conn_id: str | None = None,
    ) -> None:
        super().__init__(
            serialized_fields={"provisioned_model_id": provisioned_model_id},
            waiter_name="provisioned_model_throughput_complete",
            waiter_args={"provisionedModelId": provisioned_model_id},
            failure_message="Bedrock provisioned throughput job failed.",
            status_message="Status of Bedrock provisioned throughput job is",
            status_queries=["status"],
            return_key="provisioned_model_id",
            return_value=provisioned_model_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )

    def hook(self) -> AwsGenericHook:
        return BedrockHook(aws_conn_id=self.aws_conn_id)


class BedrockIngestionJobTrigger(AwsBaseWaiterTrigger):
    """
    Trigger when a Bedrock ingestion job reaches the COMPLETE state.

    :param knowledge_base_id: The unique identifier of the knowledge base for which to get information.
    :param data_source_id: The unique identifier of the data source in the ingestion job.
    :param ingestion_job_id: The unique identifier of the ingestion job.

    :param waiter_delay: The amount of time in seconds to wait between attempts. (default: 60)
    :param waiter_max_attempts: The maximum number of attempts to be made. (default: 10)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        *,
        knowledge_base_id: str,
        data_source_id: str,
        ingestion_job_id: str,
        waiter_delay: int = 60,
        waiter_max_attempts: int = 10,
        aws_conn_id: str | None = None,
    ) -> None:
        super().__init__(
            serialized_fields={
                "knowledge_base_id": knowledge_base_id,
                "data_source_id": data_source_id,
                "ingestion_job_id": ingestion_job_id,
            },
            waiter_name="ingestion_job_complete",
            waiter_args={
                "knowledgeBaseId": knowledge_base_id,
                "dataSourceId": data_source_id,
                "ingestionJobId": ingestion_job_id,
            },
            failure_message="Bedrock ingestion job creation failed.",
            status_message="Status of Bedrock ingestion job is",
            status_queries=["status"],
            return_key="ingestion_job_id",
            return_value=ingestion_job_id,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )

    def hook(self) -> AwsGenericHook:
        return BedrockAgentHook(aws_conn_id=self.aws_conn_id)


class BedrockBaseBatchInferenceTrigger(AwsBaseWaiterTrigger):
    """
    Trigger when a batch inference job is complete.

    :param job_arn: The Amazon Resource Name (ARN) of the batch inference job.

    :param waiter_delay: The amount of time in seconds to wait between attempts. (default: 120)
    :param waiter_max_attempts: The maximum number of attempts to be made. (default: 75)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        *,
        job_arn: str,
        waiter_name: str | ArgNotSet = NOTSET,  # This must be defined in the child class.
        waiter_delay: int = 120,
        waiter_max_attempts: int = 75,
        aws_conn_id: str | None = None,
    ) -> None:
        if waiter_name == NOTSET:
            raise NotImplementedError("Triggers must provide a waiter name.")

        super().__init__(
            serialized_fields={"job_arn": job_arn},
            waiter_name=str(waiter_name),  # Cast a string to a string to make mypy happy
            waiter_args={"jobIdentifier": job_arn},
            failure_message="Bedrock batch inference job failed.",
            status_message="Status of Bedrock batch inference job is",
            status_queries=["status"],
            return_key="job_arn",
            return_value=job_arn,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )

    def hook(self) -> AwsGenericHook:
        return BedrockHook(aws_conn_id=self.aws_conn_id)


class BedrockBatchInferenceCompletedTrigger(BedrockBaseBatchInferenceTrigger):
    """
    Trigger when a batch inference job is complete.

    :param job_arn: The Amazon Resource Name (ARN) of the batch inference job.

    :param waiter_delay: The amount of time in seconds to wait between attempts. (default: 120)
    :param waiter_max_attempts: The maximum number of attempts to be made. (default: 75)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        *,
        job_arn: str,
        waiter_delay: int = 120,
        waiter_max_attempts: int = 75,
        aws_conn_id: str | None = None,
    ) -> None:
        super().__init__(
            waiter_name="batch_inference_complete",
            job_arn=job_arn,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )


class BedrockBatchInferenceScheduledTrigger(BedrockBaseBatchInferenceTrigger):
    """
    Trigger when a batch inference job is scheduled.

    :param job_arn: The Amazon Resource Name (ARN) of the batch inference job.

    :param waiter_delay: The amount of time in seconds to wait between attempts. (default: 120)
    :param waiter_max_attempts: The maximum number of attempts to be made. (default: 75)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
    """

    def __init__(
        self,
        *,
        job_arn: str,
        waiter_delay: int = 120,
        waiter_max_attempts: int = 75,
        aws_conn_id: str | None = None,
    ) -> None:
        super().__init__(
            waiter_name="batch_inference_scheduled",
            job_arn=job_arn,
            waiter_delay=waiter_delay,
            waiter_max_attempts=waiter_max_attempts,
            aws_conn_id=aws_conn_id,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
import urllib.parse
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from botocore.exceptions import ClientError

from airflow.providers.amazon.aws.hooks.glue import GlueDataQualityHook, GlueJobHook
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.amazon.aws.links.glue import GlueJobRunDetailsLink
from airflow.providers.amazon.aws.operators.base_aws import AwsBaseOperator
from airflow.providers.amazon.aws.triggers.glue import (
    GlueDataQualityRuleRecommendationRunCompleteTrigger,
    GlueDataQualityRuleSetEvaluationRunCompleteTrigger,
    GlueJobCompleteTrigger,
)
from airflow.providers.amazon.aws.utils import validate_execute_complete_event
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.common.compat.sdk import AirflowException, conf

if TYPE_CHECKING:
    from airflow.sdk import Context


class GlueJobOperator(AwsBaseOperator[GlueJobHook]):
    """
    Create an AWS Glue Job.

    AWS Glue is a serverless Spark ETL service for running Spark Jobs on the AWS
    cloud. Language support: Python and Scala.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GlueJobOperator`

    :param job_name: unique job name per AWS Account
    :param script_location: location of ETL script. Must be a local or S3 path
    :param job_desc: job description details
    :param concurrent_run_limit: The maximum number of concurrent runs allowed for a job
    :param script_args: etl script arguments and AWS Glue arguments (templated)
    :param retry_limit: The maximum number of times to retry this job if it fails
    :param num_of_dpus: Number of AWS Glue DPUs to allocate to this Job.
    :param s3_bucket: S3 bucket where logs and local etl script will be uploaded
    :param iam_role_name: AWS IAM Role for Glue Job Execution. If set `iam_role_arn` must equal None.
    :param iam_role_arn: AWS IAM ARN for Glue Job Execution. If set `iam_role_name` must equal None.
    :param create_job_kwargs: Extra arguments for Glue Job Creation
    :param run_job_kwargs: Extra arguments for Glue Job Run
    :param wait_for_completion: Whether to wait for job run completion. (default: True)
    :param deferrable: If True, the operator will wait asynchronously for the job to complete.
        This implies waiting for completion. This mode requires aiobotocore module to be installed.
        (default: False)
    :param verbose: If True, Glue Job Run logs show in the Airflow Task Logs.  (default: False)
    :param update_config: If True, Operator will update job configuration.  (default: False)
    :param replace_script_file: If True, the script file will be replaced in S3. (default: False)
    :param stop_job_run_on_kill: If True, Operator will stop the job run when task is killed.
    :param sleep_before_return: time in seconds to wait before returning final status. This is meaningful in case
        of limiting concurrency, Glue needs 5-10 seconds to clean up resources.
        Thus if status is returned immediately it might end up in case of more than 1 concurrent run.
        It is recommended to set this parameter to 10 when you are using concurrency=1.
        For more information see:
        https://repost.aws/questions/QUaKgpLBMPSGWO0iq2Fob_bw/glue-run-concurrent-jobs#ANFpCL2fRnQRqgDFuIU_rpvA
    :param waiter_delay: Time in seconds to wait between status checks. (default: 60)
    :param waiter_max_attempts: Maximum number of attempts to check for job completion. (default: 20)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = GlueJobHook
    template_fields: Sequence[str] = aws_template_fields(
        "job_name",
        "script_location",
        "script_args",
        "create_job_kwargs",
        "run_job_kwargs",
        "s3_bucket",
        "iam_role_name",
        "iam_role_arn",
    )
    template_ext: Sequence[str] = ()
    template_fields_renderers = {
        "script_args": "json",
        "create_job_kwargs": "json",
        "run_job_kwargs": "json",
    }
    ui_color = "#ededed"

    operator_extra_links = (GlueJobRunDetailsLink(),)

    def __init__(
        self,
        *,
        job_name: str = "aws_glue_default_job",
        job_desc: str = "AWS Glue Job with Airflow",
        script_location: str | None = None,
        concurrent_run_limit: int | None = None,
        script_args: dict | None = None,
        retry_limit: int = 0,
        num_of_dpus: int | float | None = None,
        s3_bucket: str | None = None,
        iam_role_name: str | None = None,
        iam_role_arn: str | None = None,
        create_job_kwargs: dict | None = None,
        run_job_kwargs: dict | None = None,
        wait_for_completion: bool = True,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        verbose: bool = False,
        replace_script_file: bool = False,
        update_config: bool = False,
        stop_job_run_on_kill: bool = False,
        sleep_before_return: int = 0,
        job_poll_interval: int | float = 6,
        waiter_delay: int = 60,
        waiter_max_attempts: int = 75,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.job_name = job_name
        self.job_desc = job_desc
        self.script_location = script_location
        self.concurrent_run_limit = concurrent_run_limit or 1
        self.script_args = script_args or {}
        self.retry_limit = retry_limit
        self.num_of_dpus = num_of_dpus
        self.s3_bucket = s3_bucket
        self.iam_role_name = iam_role_name
        self.iam_role_arn = iam_role_arn
        self.s3_protocol = "s3://"
        self.s3_artifacts_prefix = "artifacts/glue-scripts/"
        self.create_job_kwargs = create_job_kwargs
        self.run_job_kwargs = run_job_kwargs or {}
        self.wait_for_completion = wait_for_completion
        self.verbose = verbose
        self.update_config = update_config
        self.replace_script_file = replace_script_file
        self.deferrable = deferrable
        self.job_poll_interval = job_poll_interval
        self.stop_job_run_on_kill = stop_job_run_on_kill
        self._job_run_id: str | None = None
        self.sleep_before_return: int = sleep_before_return
        self.s3_script_location: str | None = None
        self.waiter_delay = waiter_delay
        self.waiter_max_attempts = waiter_max_attempts

    @property
    def _hook_parameters(self):
        # Upload script to S3 before creating the hook.
        if self.script_location is None:
            self.s3_script_location = None
        # location provided, but it's not in S3 yet.
        elif self.script_location and self.s3_script_location is None:
            if not self.script_location.startswith(self.s3_protocol):
                self.upload_etl_script_to_s3()
            else:
                self.s3_script_location = self.script_location

        return {
            **super()._hook_parameters,
            "job_name": self.job_name,
            "desc": self.job_desc,
            "concurrent_run_limit": self.concurrent_run_limit,
            "script_location": self.s3_script_location,
            "retry_limit": self.retry_limit,
            "num_of_dpus": self.num_of_dpus,
            "aws_conn_id": self.aws_conn_id,
            "region_name": self.region_name,
            "s3_bucket": self.s3_bucket,
            "iam_role_name": self.iam_role_name,
            "iam_role_arn": self.iam_role_arn,
            "create_job_kwargs": self.create_job_kwargs,
            "update_config": self.update_config,
            "job_poll_interval": self.job_poll_interval,
        }

    def upload_etl_script_to_s3(self):
        """Upload the ETL script to S3."""
        s3_hook = S3Hook(aws_conn_id=self.aws_conn_id)
        script_name = os.path.basename(self.script_location)
        s3_hook.load_file(
            self.script_location,
            self.s3_artifacts_prefix + script_name,
            bucket_name=self.s3_bucket,
            replace=self.replace_script_file,
        )
        self.s3_script_location = f"s3://{self.s3_bucket}/{self.s3_artifacts_prefix}{script_name}"

    def execute(self, context: Context):
        """
        Execute AWS Glue Job from Airflow.

        :return: the current Glue job ID.
        """
        self.log.info(
            "Initializing AWS Glue Job: %s. Wait for completion: %s",
            self.job_name,
            self.wait_for_completion,
        )
        glue_job_run = self.hook.initialize_job(self.script_args, self.run_job_kwargs)
        self._job_run_id = glue_job_run["JobRunId"]
        glue_job_run_url = GlueJobRunDetailsLink.format_str.format(
            aws_domain=GlueJobRunDetailsLink.get_aws_domain(self.hook.conn_partition),
            region_name=self.hook.conn_region_name,
            job_name=urllib.parse.quote(self.job_name, safe=""),
            job_run_id=self._job_run_id,
        )
        GlueJobRunDetailsLink.persist(
            context=context,
            operator=self,
            region_name=self.hook.conn_region_name,
            aws_partition=self.hook.conn_partition,
            job_name=urllib.parse.quote(self.job_name, safe=""),
            job_run_id=self._job_run_id,
        )
        self.log.info("You can monitor this Glue Job run at: %s", glue_job_run_url)

        if self.deferrable:
            self.defer(
                trigger=GlueJobCompleteTrigger(
                    job_name=self.job_name,
                    run_id=self._job_run_id,
                    verbose=self.verbose,
                    aws_conn_id=self.aws_conn_id,
                    waiter_delay=self.waiter_delay,
                    waiter_max_attempts=self.waiter_max_attempts,
                    region_name=self.region_name,
                ),
                method_name="execute_complete",
            )
        elif self.wait_for_completion:
            glue_job_run = self.hook.job_completion(
                self.job_name, self._job_run_id, self.verbose, self.sleep_before_return
            )
            self.log.info(
                "AWS Glue Job: %s status: %s. Run Id: %s",
                self.job_name,
                glue_job_run["JobRunState"],
                self._job_run_id,
            )
        else:
            self.log.info("AWS Glue Job: %s. Run Id: %s", self.job_name, self._job_run_id)
        return self._job_run_id

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> str:
        validated_event = validate_execute_complete_event(event)

        if validated_event["status"] != "success":
            raise AirflowException(f"Error in glue job: {validated_event}")
        return validated_event["run_id"]

    def on_kill(self):
        """Cancel the running AWS Glue Job."""
        if self.stop_job_run_on_kill:
            self.log.info("Stopping AWS Glue Job: %s. Run Id: %s", self.job_name, self._job_run_id)
            response = self.hook.conn.batch_stop_job_run(
                JobName=self.job_name,
                JobRunIds=[self._job_run_id],
            )
            if not response["SuccessfulSubmissions"]:
                self.log.error("Failed to stop AWS Glue Job: %s. Run Id: %s", self.job_name, self._job_run_id)


class GlueDataQualityOperator(AwsBaseOperator[GlueDataQualityHook]):
    """
    Creates a data quality ruleset with DQDL rules applied to a specified Glue table.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GlueDataQualityOperator`

    :param name: A unique name for the data quality ruleset.
    :param ruleset: A Data Quality Definition Language (DQDL) ruleset.
        For more information, see the Glue developer guide.
    :param description: A description of the data quality ruleset.
    :param update_rule_set: To update existing ruleset, Set this flag to True. (default: False)
    :param data_quality_ruleset_kwargs: Extra arguments for RuleSet.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = GlueDataQualityHook
    template_fields: Sequence[str] = aws_template_fields(
        "name", "ruleset", "description", "data_quality_ruleset_kwargs"
    )

    template_fields_renderers = {
        "data_quality_ruleset_kwargs": "json",
    }
    ui_color = "#ededed"

    def __init__(
        self,
        *,
        name: str,
        ruleset: str,
        description: str = "AWS Glue Data Quality Rule Set With Airflow",
        update_rule_set: bool = False,
        data_quality_ruleset_kwargs: dict | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.name = name
        self.ruleset = ruleset.strip()
        self.description = description
        self.update_rule_set = update_rule_set
        self.data_quality_ruleset_kwargs = data_quality_ruleset_kwargs or {}

    def validate_inputs(self) -> None:
        if not self.ruleset.startswith("Rules") or not self.ruleset.endswith("]"):
            raise AttributeError("RuleSet must starts with Rules = [ and ends with ]")

        if self.data_quality_ruleset_kwargs.get("TargetTable"):
            target_table = self.data_quality_ruleset_kwargs["TargetTable"]

            if not target_table.get("TableName") or not target_table.get("DatabaseName"):
                raise AttributeError("Target table must have DatabaseName and TableName")

    def execute(self, context: Context):
        self.validate_inputs()

        config = {
            "Name": self.name,
            "Ruleset": self.ruleset,
            "Description": self.description,
            **self.data_quality_ruleset_kwargs,
        }
        try:
            if self.update_rule_set:
                self.hook.conn.update_data_quality_ruleset(**config)
                self.log.info("AWS Glue data quality ruleset updated successfully")
            else:
                self.hook.conn.create_data_quality_ruleset(**config)
                self.log.info("AWS Glue data quality ruleset created successfully")
        except ClientError as error:
            raise AirflowException(
                f"AWS Glue data quality ruleset failed: {error.response['Error']['Message']}"
            )


class GlueDataQualityRuleSetEvaluationRunOperator(AwsBaseOperator[GlueDataQualityHook]):
    """
    Evaluate a ruleset against a data source (Glue table).

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GlueDataQualityRuleSetEvaluationRunOperator`

    :param datasource: The data source (Glue table) associated with this run. (templated)
    :param role: IAM role supplied for job execution. (templated)
    :param rule_set_names: A list of ruleset names for evaluation. (templated)
    :param number_of_workers: The number of G.1X workers to be used in the run. (default: 5)
    :param timeout: The timeout for a run in minutes. This is the maximum time that a run can consume resources
        before it is terminated and enters TIMEOUT status. (default: 2,880)
    :param verify_result_status: Validate all the ruleset rules evaluation run results,
        If any of the rule status is Fail or Error then an exception is thrown. (default: True)
    :param show_results: Displays all the ruleset rules evaluation run results. (default: True)
    :param rule_set_evaluation_run_kwargs: Extra arguments for evaluation run. (templated)
    :param wait_for_completion: Whether to wait for job to stop. (default: True)
    :param waiter_delay: Time in seconds to wait between status checks. (default: 60)
    :param waiter_max_attempts: Maximum number of attempts to check for job completion. (default: 20)
    :param deferrable: If True, the operator will wait asynchronously for the job to stop.
        This implies waiting for completion. This mode requires aiobotocore module to be installed.
        (default: False)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = GlueDataQualityHook

    template_fields: Sequence[str] = aws_template_fields(
        "datasource",
        "role",
        "rule_set_names",
        "rule_set_evaluation_run_kwargs",
    )

    template_fields_renderers = {"datasource": "json", "rule_set_evaluation_run_kwargs": "json"}

    ui_color = "#ededed"

    def __init__(
        self,
        *,
        datasource: dict,
        role: str,
        rule_set_names: list[str],
        number_of_workers: int = 5,
        timeout: int = 2880,
        verify_result_status: bool = True,
        show_results: bool = True,
        rule_set_evaluation_run_kwargs: dict[str, Any] | None = None,
        wait_for_completion: bool = True,
        waiter_delay: int = 60,
        waiter_max_attempts: int = 20,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.datasource = datasource
        self.role = role
        self.rule_set_names = rule_set_names
        self.number_of_workers = number_of_workers
        self.timeout = timeout
        self.verify_result_status = verify_result_status
        self.show_results = show_results
        self.rule_set_evaluation_run_kwargs = rule_set_evaluation_run_kwargs or {}
        self.wait_for_completion = wait_for_completion
        self.waiter_delay = waiter_delay
        self.waiter_max_attempts = waiter_max_attempts
        self.deferrable = deferrable

    def validate_inputs(self) -> None:
        glue_table = self.datasource.get("GlueTable", {})

        if not glue_table.get("DatabaseName") or not glue_table.get("TableName"):
            raise AttributeError("DataSource glue table must have DatabaseName and TableName")

        not_found_ruleset = [
            ruleset_name
            for ruleset_name in self.rule_set_names
            if not self.hook.has_data_quality_ruleset(ruleset_name)
        ]

        if not_found_ruleset:
            raise AirflowException(f"Following RulesetNames are not found {not_found_ruleset}")

    def execute(self, context: Context) -> str:
        self.validate_inputs()

        self.log.info(
            "Submitting AWS Glue data quality ruleset evaluation run for RulesetNames %s", self.rule_set_names
        )

        response = self.hook.conn.start_data_quality_ruleset_evaluation_run(
            DataSource=self.datasource,
            Role=self.role,
            NumberOfWorkers=self.number_of_workers,
            Timeout=self.timeout,
            RulesetNames=self.rule_set_names,
            **self.rule_set_evaluation_run_kwargs,
        )

        evaluation_run_id = response["RunId"]

        message_description = (
            f"AWS Glue data quality ruleset evaluation run RunId: {evaluation_run_id} to complete."
        )
        if self.deferrable:
            self.log.info("Deferring %s", message_description)
            self.defer(
                trigger=GlueDataQualityRuleSetEvaluationRunCompleteTrigger(
                    evaluation_run_id=response["RunId"],
                    waiter_delay=self.waiter_delay,
                    waiter_max_attempts=self.waiter_max_attempts,
                    aws_conn_id=self.aws_conn_id,
                ),
                method_name="execute_complete",
            )

        elif self.wait_for_completion:
            self.log.info("Waiting for %s", message_description)

            self.hook.get_waiter("data_quality_ruleset_evaluation_run_complete").wait(
                RunId=evaluation_run_id,
                WaiterConfig={"Delay": self.waiter_delay, "MaxAttempts": self.waiter_max_attempts},
            )

            self.log.info(
                "AWS Glue data quality ruleset evaluation run completed RunId: %s", evaluation_run_id
            )

            self.hook.validate_evaluation_run_results(
                evaluation_run_id=evaluation_run_id,
                show_results=self.show_results,
                verify_result_status=self.verify_result_status,
            )
        else:
            self.log.info("AWS Glue data quality ruleset evaluation run runId: %s.", evaluation_run_id)

        return evaluation_run_id

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> str:
        validated_event = validate_execute_complete_event(event)

        if validated_event["status"] != "success":
            raise AirflowException(f"Error: AWS Glue data quality ruleset evaluation run: {validated_event}")

        self.hook.validate_evaluation_run_results(
            evaluation_run_id=validated_event["evaluation_run_id"],
            show_results=self.show_results,
            verify_result_status=self.verify_result_status,
        )

        return validated_event["evaluation_run_id"]


class GlueDataQualityRuleRecommendationRunOperator(AwsBaseOperator[GlueDataQualityHook]):
    """
    Starts a recommendation run that is used to generate rules, Glue Data Quality analyzes the data and comes up with recommendations for a potential ruleset.

    Recommendation runs are automatically deleted after 90 days.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GlueDataQualityRuleRecommendationRunOperator`

    :param datasource: The data source (Glue table) associated with this run. (templated)
    :param role: IAM role supplied for job execution. (templated)
    :param number_of_workers: The number of G.1X workers to be used in the run. (default: 5)
    :param timeout: The timeout for a run in minutes. This is the maximum time that a run can consume resources
        before it is terminated and enters TIMEOUT status. (default: 2,880)
    :param show_results: Displays the recommended ruleset (a set of rules), when recommendation run completes. (default: True)
    :param recommendation_run_kwargs: Extra arguments for recommendation run. (templated)
    :param wait_for_completion: Whether to wait for job to stop. (default: True)
    :param waiter_delay: Time in seconds to wait between status checks. (default: 60)
    :param waiter_max_attempts: Maximum number of attempts to check for job completion. (default: 20)
    :param deferrable: If True, the operator will wait asynchronously for the job to stop.
        This implies waiting for completion. This mode requires aiobotocore module to be installed.
        (default: False)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = GlueDataQualityHook
    template_fields: Sequence[str] = aws_template_fields(
        "datasource",
        "role",
        "recommendation_run_kwargs",
    )

    template_fields_renderers = {"datasource": "json", "recommendation_run_kwargs": "json"}

    ui_color = "#ededed"

    def __init__(
        self,
        *,
        datasource: dict,
        role: str,
        number_of_workers: int = 5,
        timeout: int = 2880,
        show_results: bool = True,
        recommendation_run_kwargs: dict[str, Any] | None = None,
        wait_for_completion: bool = True,
        waiter_delay: int = 60,
        waiter_max_attempts: int = 20,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.datasource = datasource
        self.role = role
        self.number_of_workers = number_of_workers
        self.timeout = timeout
        self.show_results = show_results
        self.recommendation_run_kwargs = recommendation_run_kwargs or {}
        self.wait_for_completion = wait_for_completion
        self.waiter_delay = waiter_delay
        self.waiter_max_attempts = waiter_max_attempts
        self.deferrable = deferrable

    def execute(self, context: Context) -> str:
        glue_table = self.datasource.get("GlueTable", {})

        if not glue_table.get("DatabaseName") or not glue_table.get("TableName"):
            raise AttributeError("DataSource glue table must have DatabaseName and TableName")

        self.log.info("Submitting AWS Glue data quality recommendation run with %s", self.datasource)

        try:
            response = self.hook.conn.start_data_quality_rule_recommendation_run(
                DataSource=self.datasource,
                Role=self.role,
                NumberOfWorkers=self.number_of_workers,
                Timeout=self.timeout,
                **self.recommendation_run_kwargs,
            )
        except ClientError as error:
            raise AirflowException(
                f"AWS Glue data quality recommendation run failed: {error.response['Error']['Message']}"
            )

        recommendation_run_id = response["RunId"]

        message_description = (
            f"AWS Glue data quality recommendation run RunId: {recommendation_run_id} to complete."
        )
        if self.deferrable:
            self.log.info("Deferring %s", message_description)
            self.defer(
                trigger=GlueDataQualityRuleRecommendationRunCompleteTrigger(
                    recommendation_run_id=recommendation_run_id,
                    waiter_delay=self.waiter_delay,
                    waiter_max_attempts=self.waiter_max_attempts,
                    aws_conn_id=self.aws_conn_id,
                ),
                method_name="execute_complete",
            )

        elif self.wait_for_completion:
            self.log.info("Waiting for %s", message_description)

            self.hook.get_waiter("data_quality_rule_recommendation_run_complete").wait(
                RunId=recommendation_run_id,
                WaiterConfig={"Delay": self.waiter_delay, "MaxAttempts": self.waiter_max_attempts},
            )
            self.log.info(
                "AWS Glue data quality recommendation run completed RunId: %s", recommendation_run_id
            )

            if self.show_results:
                self.hook.log_recommendation_results(run_id=recommendation_run_id)

        else:
            self.log.info(message_description)

        return recommendation_run_id

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> str:
        validated_event = validate_execute_complete_event(event)

        if validated_event["status"] != "success":
            raise AirflowException(f"Error: AWS Glue data quality rule recommendation run: {validated_event}")

        if self.show_results:
            self.hook.log_recommendation_results(run_id=validated_event["recommendation_run_id"])

        return validated_event["recommendation_run_id"]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest import mock

import boto3
import botocore
import pytest

from airflow.providers.amazon.aws.hooks.neptune import NeptuneHook


class TestCustomNeptuneWaiters:
    """Test waiters from ``amazon/aws/waiters/neptune.json``."""

    @pytest.fixture(autouse=True)
    def _setup_test_cases(self, monkeypatch):
        self.client = boto3.client("neptune", region_name="eu-west-3")
        monkeypatch.setattr(NeptuneHook, "conn", self.client)

    def test_service_waiters(self):
        hook_waiters = NeptuneHook(aws_conn_id=None).list_waiters()
        assert "cluster_available" in hook_waiters

    @pytest.fixture
    def mock_describe_clusters(self):
        with mock.patch.object(self.client, "describe_db_clusters") as m:
            yield m

    @staticmethod
    def get_status_response(status):
        return {"DBClusters": [{"Status": status}]}

    def test_cluster_available(self, mock_describe_clusters):
        mock_describe_clusters.return_value = {"DBClusters": [{"Status": "available"}]}
        waiter = NeptuneHook(aws_conn_id=None).get_waiter("cluster_available")
        waiter.wait(DBClusterIdentifier="test_cluster")

    def test_cluster_available_failed(self, mock_describe_clusters):
        mock_describe_clusters.return_value = {"DBClusters": [{"Status": "migration-failed"}]}
        waiter = NeptuneHook(aws_conn_id=None).get_waiter("cluster_available")
        with pytest.raises(botocore.exceptions.WaiterError):
            waiter.wait(DBClusterIdentifier="test_cluster")

    def test_starting_up(self, mock_describe_clusters):
        """Test job succeeded"""
        mock_describe_clusters.side_effect = [
            self.get_status_response("stopped"),
            self.get_status_response("starting"),
            self.get_status_response("available"),
        ]
        waiter = NeptuneHook(aws_conn_id=None).get_waiter("cluster_available")
        waiter.wait(cluster_identifier="test_cluster", WaiterConfig={"Delay": 0.2, "MaxAttempts": 4})

    def test_cluster_stopped(self, mock_describe_clusters):
        mock_describe_clusters.return_value = {"DBClusters": [{"Status": "stopped"}]}
        waiter = NeptuneHook(aws_conn_id=None).get_waiter("cluster_stopped")
        waiter.wait(DBClusterIdentifier="test_cluster")

    def test_cluster_stopped_failed(self, mock_describe_clusters):
        mock_describe_clusters.return_value = {"DBClusters": [{"Status": "migration-failed"}]}
        waiter = NeptuneHook(aws_conn_id=None).get_waiter("cluster_stopped")
        with pytest.raises(botocore.exceptions.WaiterError):
            waiter.wait(DBClusterIdentifier="test_cluster")

    def test_stopping(self, mock_describe_clusters):
        mock_describe_clusters.side_effect = [
            self.get_status_response("available"),
            self.get_status_response("stopping"),
            self.get_status_response("stopped"),
        ]
        waiter = NeptuneHook(aws_conn_id=None).get_waiter("cluster_stopped")
        waiter.wait(cluster_identifier="test_cluster", WaiterConfig={"Delay": 0.2, "MaxAttempts": 4})



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import patch

import boto3
import pytest
from moto import mock_aws

from airflow.providers.amazon.aws.sensors.cloud_formation import (
    CloudFormationCreateStackSensor,
    CloudFormationDeleteStackSensor,
)


@pytest.fixture
def mocked_hook_client():
    with patch("airflow.providers.amazon.aws.hooks.cloud_formation.CloudFormationHook.conn") as m:
        yield m


class TestCloudFormationCreateStackSensor:
    @mock_aws
    def setup_method(self, method):
        self.client = boto3.client("cloudformation", region_name="us-east-1")

    def test_init(self):
        sensor = CloudFormationCreateStackSensor(
            task_id="cf_create_stack_init",
            stack_name="fake-stack",
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="eu-central-1",
            verify=False,
            botocore_config={"read_timeout": 42},
        )
        assert sensor.hook.client_type == "cloudformation"
        assert sensor.hook.resource_type is None
        assert sensor.hook.aws_conn_id == "fake-conn-id"
        assert sensor.hook._region_name == "eu-central-1"
        assert sensor.hook._verify is False
        assert sensor.hook._config is not None
        assert sensor.hook._config.read_timeout == 42

        sensor = CloudFormationCreateStackSensor(task_id="cf_create_stack_init", stack_name="fake-stack")
        assert sensor.hook.aws_conn_id == "aws_default"
        assert sensor.hook._region_name is None
        assert sensor.hook._verify is None
        assert sensor.hook._config is None

    @mock_aws
    def test_poke(self):
        self.client.create_stack(StackName="foobar", TemplateBody='{"Resources": {}}')
        op = CloudFormationCreateStackSensor(task_id="task", stack_name="foobar")
        assert op.poke({})

    def test_poke_false(self, mocked_hook_client):
        mocked_hook_client.describe_stacks.return_value = {"Stacks": [{"StackStatus": "CREATE_IN_PROGRESS"}]}
        op = CloudFormationCreateStackSensor(task_id="task", stack_name="foo")
        assert not op.poke({})

    def test_poke_stack_in_unsuccessful_state(self, mocked_hook_client):
        mocked_hook_client.describe_stacks.return_value = {"Stacks": [{"StackStatus": "bar"}]}
        op = CloudFormationCreateStackSensor(task_id="task", stack_name="foo")
        with pytest.raises(ValueError, match="Stack foo in bad state: bar"):
            op.poke({})


class TestCloudFormationDeleteStackSensor:
    @mock_aws
    def setup_method(self, method):
        self.client = boto3.client("cloudformation", region_name="us-east-1")

    def test_init(self):
        sensor = CloudFormationDeleteStackSensor(
            task_id="cf_delete_stack_init",
            stack_name="fake-stack",
            # Generic hooks parameters
            aws_conn_id="fake-conn-id",
            region_name="ca-west-1",
            verify=True,
            botocore_config={"read_timeout": 42},
        )
        assert sensor.hook.client_type == "cloudformation"
        assert sensor.hook.resource_type is None
        assert sensor.hook.aws_conn_id == "fake-conn-id"
        assert sensor.hook._region_name == "ca-west-1"
        assert sensor.hook._verify is True
        assert sensor.hook._config is not None
        assert sensor.hook._config.read_timeout == 42

        sensor = CloudFormationDeleteStackSensor(task_id="cf_delete_stack_init", stack_name="fake-stack")
        assert sensor.hook.aws_conn_id == "aws_default"
        assert sensor.hook._region_name is None
        assert sensor.hook._verify is None
        assert sensor.hook._config is None

    @mock_aws
    def test_poke(self):
        stack_name = "foobar"
        self.client.create_stack(StackName=stack_name, TemplateBody='{"Resources": {}}')
        self.client.delete_stack(StackName=stack_name)
        op = CloudFormationDeleteStackSensor(task_id="task", stack_name=stack_name)
        assert op.poke({})

    def test_poke_false(self, mocked_hook_client):
        mocked_hook_client.describe_stacks.return_value = {"Stacks": [{"StackStatus": "DELETE_IN_PROGRESS"}]}
        op = CloudFormationDeleteStackSensor(task_id="task", stack_name="foo")
        assert not op.poke({})

    def test_poke_stack_in_unsuccessful_state(self, mocked_hook_client):
        mocked_hook_client.describe_stacks.return_value = {"Stacks": [{"StackStatus": "bar"}]}
        op = CloudFormationDeleteStackSensor(task_id="task", stack_name="foo")
        with pytest.raises(ValueError, match="Stack foo in bad state: bar"):
            op.poke({})

    @mock_aws
    def test_poke_stack_does_not_exist(self):
        op = CloudFormationDeleteStackSensor(task_id="task", stack_name="foo")
        assert op.poke({})



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from datetime import datetime

from airflow.providers.amazon.aws.executors.ecs.boto_schema import (
    BotoContainerSchema,
    BotoDescribeTasksSchema,
    BotoFailureSchema,
    BotoRunTaskSchema,
    BotoTaskSchema,
)
from airflow.providers.amazon.aws.executors.ecs.utils import EcsExecutorTask


class TestBotoSchema:
    def test_boto_container_schema_load(self):
        schema = BotoContainerSchema()
        data = {
            "exitCode": 0,
            "lastStatus": "STOPPED",
            "name": "test_container",
            "reason": "Essential container in task exited",
            "containerArn": "arn:aws:ecs:us-east-1:123456789012:container/test-cluster/1234567890abcdef0",
        }
        result = schema.load(data)
        assert result["exit_code"] == 0
        assert result["last_status"] == "STOPPED"
        assert result["name"] == "test_container"
        assert result["reason"] == "Essential container in task exited"
        assert (
            result["container_arn"]
            == "arn:aws:ecs:us-east-1:123456789012:container/test-cluster/1234567890abcdef0"
        )

    def test_boto_container_schema_load_minimal(self):
        schema = BotoContainerSchema()
        data = {"name": "minimal_container"}
        result = schema.load(data)
        assert result["name"] == "minimal_container"
        assert result.get("exit_code") is None
        assert result.get("last_status") is None
        assert result.get("reason") is None
        assert result.get("container_arn") is None

    def test_boto_container_schema_exclude_unknown(self):
        schema = BotoContainerSchema()
        data = {"name": "test_container", "unknownField": "should_be_ignored"}
        result = schema.load(data)
        assert "unknownField" not in result

    def test_boto_task_schema_load(self):
        schema = BotoTaskSchema()
        container_data = {
            "exitCode": 0,
            "lastStatus": "STOPPED",
            "name": "test_container",
            "reason": "Essential container in task exited",
            "containerArn": "arn:aws:ecs:us-east-1:123456789012:container/test-cluster/1234567890abcdef0",
        }
        data = {
            "taskArn": "arn:aws:ecs:us-east-1:123456789012:task/test-cluster/1234567890abcdef0",
            "lastStatus": "STOPPED",
            "desiredStatus": "STOPPED",
            "containers": [container_data],
            "startedAt": datetime(2023, 1, 1),
            "stoppedReason": "Task failed to start",
        }
        result = schema.load(data)
        assert isinstance(result, EcsExecutorTask)
        assert result.task_arn == "arn:aws:ecs:us-east-1:123456789012:task/test-cluster/1234567890abcdef0"
        assert result.last_status == "STOPPED"
        assert result.desired_status == "STOPPED"
        assert len(result.containers) == 1
        assert result.containers[0]["name"] == "test_container"
        assert result.started_at == datetime(2023, 1, 1)
        assert result.stopped_reason == "Task failed to start"

    def test_boto_task_schema_load_minimal(self):
        schema = BotoTaskSchema()
        container_data = {"name": "minimal_container_in_task"}
        data = {
            "taskArn": "arn:aws:ecs:us-east-1:123456789012:task/test-cluster/1234567890abcdef0",
            "lastStatus": "RUNNING",
            "desiredStatus": "RUNNING",
            "containers": [container_data],
        }
        result = schema.load(data)
        assert isinstance(result, EcsExecutorTask)
        assert result.task_arn == "arn:aws:ecs:us-east-1:123456789012:task/test-cluster/1234567890abcdef0"
        assert result.last_status == "RUNNING"
        assert result.desired_status == "RUNNING"
        assert len(result.containers) == 1
        assert result.containers[0]["name"] == "minimal_container_in_task"
        assert result.started_at is None
        assert result.stopped_reason is None

    def test_boto_task_schema_exclude_unknown(self):
        schema = BotoTaskSchema()
        container_data = {"name": "test_container"}
        data = {
            "taskArn": "arn:aws:ecs:us-east-1:123456789012:task/test-cluster/1234567890abcdef0",
            "lastStatus": "RUNNING",
            "desiredStatus": "RUNNING",
            "containers": [container_data],
            "unknownTaskField": "should_be_ignored",
        }
        result = schema.load(data)
        # EcsExecutorTask doesn't store unknown fields, so we check the deserialized dict before object creation
        # by checking the raw data passed to EcsExecutorTask constructor if possible,
        # or simply ensure no error occurs and the object is created.
        # A more direct way would be to mock EcsExecutorTask and inspect its kwargs.
        # For now, we just ensure it loads without error and produces the correct type.
        assert isinstance(result, EcsExecutorTask)

    def test_boto_failure_schema_load(self):
        schema = BotoFailureSchema()
        data = {
            "arn": "arn:aws:ecs:us-east-1:123456789012:container/test-cluster/1234567890abcdef0",
            "reason": "MISSING",
        }
        result = schema.load(data)
        assert result["arn"] == "arn:aws:ecs:us-east-1:123456789012:container/test-cluster/1234567890abcdef0"
        assert result["reason"] == "MISSING"

    def test_boto_failure_schema_load_minimal(self):
        schema = BotoFailureSchema()
        data = {}
        result = schema.load(data)
        assert result.get("arn") is None
        assert result.get("reason") is None

    def test_boto_failure_schema_exclude_unknown(self):
        schema = BotoFailureSchema()
        data = {"arn": "test_arn", "unknownField": "should_be_ignored"}
        result = schema.load(data)
        assert "unknownField" not in result

    def test_boto_run_task_schema_load(self):
        schema = BotoRunTaskSchema()
        task_data = {
            "taskArn": "arn:aws:ecs:us-east-1:123456789012:task/test-cluster/1234567890abcdef0",
            "lastStatus": "RUNNING",
            "desiredStatus": "RUNNING",
            "containers": [{"name": "test_container"}],
        }
        failure_data = {
            "arn": "arn:aws:ecs:us-east-1:123456789012:container/test-cluster/badabcdef0",
            "reason": "MISSING",
        }
        data = {"tasks": [task_data], "failures": [failure_data]}
        result = schema.load(data)

        assert len(result["tasks"]) == 1
        assert isinstance(result["tasks"][0], EcsExecutorTask)
        assert (
            result["tasks"][0].task_arn
            == "arn:aws:ecs:us-east-1:123456789012:task/test-cluster/1234567890abcdef0"
        )

        assert len(result["failures"]) == 1
        assert (
            result["failures"][0]["arn"]
            == "arn:aws:ecs:us-east-1:123456789012:container/test-cluster/badabcdef0"
        )
        assert result["failures"][0]["reason"] == "MISSING"

    def test_boto_run_task_schema_exclude_unknown(self):
        schema = BotoRunTaskSchema()
        data = {
            "tasks": [],
            "failures": [],
            "unknownRunTaskField": "should_be_ignored",
        }
        result = schema.load(data)
        assert "unknownRunTaskField" not in result

    def test_boto_describe_tasks_schema_load(self):
        schema = BotoDescribeTasksSchema()
        task_data = {
            "taskArn": "arn:aws:ecs:us-east-1:123456789012:task/test-cluster/1234567890abcdef1",
            "lastStatus": "STOPPED",
            "desiredStatus": "STOPPED",
            "containers": [{"name": "another_container", "exitCode": 1}],
        }
        failure_data = {
            "arn": "arn:aws:ecs:us-east-1:123456789012:container/test-cluster/anotherbad",
            "reason": "UNABLE",
        }
        data = {"tasks": [task_data], "failures": [failure_data]}
        result = schema.load(data)

        assert len(result["tasks"]) == 1
        assert isinstance(result["tasks"][0], EcsExecutorTask)
        assert (
            result["tasks"][0].task_arn
            == "arn:aws:ecs:us-east-1:123456789012:task/test-cluster/1234567890abcdef1"
        )
        assert result["tasks"][0].containers[0]["exit_code"] == 1

        assert len(result["failures"]) == 1
        assert (
            result["failures"][0]["arn"]
            == "arn:aws:ecs:us-east-1:123456789012:container/test-cluster/anotherbad"
        )
        assert result["failures"][0]["reason"] == "UNABLE"

    def test_boto_describe_tasks_schema_exclude_unknown(self):
        schema = BotoDescribeTasksSchema()
        data = {
            "tasks": [],
            "failures": [],
            "unknownDescribeTasksField": "should_be_ignored",
        }
        result = schema.load(data)
        assert "unknownDescribeTasksField" not in result



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from airflow.providers.amazon.aws.hooks.rds import RdsHook
from airflow.providers.amazon.aws.operators.rds import (
    RdsCancelExportTaskOperator,
    RdsCreateDbInstanceOperator,
    RdsCreateDbSnapshotOperator,
    RdsDeleteDbInstanceOperator,
    RdsDeleteDbSnapshotOperator,
    RdsStartExportTaskOperator,
)
from airflow.providers.amazon.aws.operators.s3 import S3CreateBucketOperator, S3DeleteBucketOperator
from airflow.providers.amazon.aws.sensors.rds import RdsExportTaskExistenceSensor, RdsSnapshotExistenceSensor

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

DAG_ID = "example_rds_export"

# Externally fetched variables:
KMS_KEY_ID_KEY = "KMS_KEY_ID"
ROLE_ARN_KEY = "ROLE_ARN"

sys_test_context_task = (
    SystemTestContextBuilder().add_variable(KMS_KEY_ID_KEY).add_variable(ROLE_ARN_KEY).build()
)


@task
def get_snapshot_arn(snapshot_name: str) -> str:
    result = RdsHook().conn.describe_db_snapshots(DBSnapshotIdentifier=snapshot_name)
    return result["DBSnapshots"][0]["DBSnapshotArn"]


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]

    bucket_name: str = f"{env_id}-bucket"

    rds_db_name: str = f"{env_id}_db"
    rds_instance_name: str = f"{env_id}-instance"
    rds_snapshot_name: str = f"{env_id}-snapshot"
    rds_export_task_id: str = f"{env_id}-export-task"

    create_bucket = S3CreateBucketOperator(
        task_id="create_bucket",
        bucket_name=bucket_name,
    )

    create_db_instance = RdsCreateDbInstanceOperator(
        task_id="create_db_instance",
        db_instance_identifier=rds_instance_name,
        db_instance_class="db.t4g.micro",
        engine="postgres",
        rds_kwargs={
            "MasterUsername": "rds_username",
            # NEVER store your production password in plaintext in a DAG like this.
            # Use Airflow Secrets or a secret manager for this in production.
            "MasterUserPassword": "rds_password",
            "AllocatedStorage": 20,
            "DBName": rds_db_name,
            "PubliclyAccessible": False,
        },
    )

    create_snapshot = RdsCreateDbSnapshotOperator(
        task_id="create_snapshot",
        db_type="instance",
        db_identifier=rds_instance_name,
        db_snapshot_identifier=rds_snapshot_name,
    )

    await_snapshot = RdsSnapshotExistenceSensor(
        task_id="snapshot_sensor",
        db_type="instance",
        db_snapshot_identifier=rds_snapshot_name,
        target_statuses=["available"],
    )

    snapshot_arn = get_snapshot_arn(rds_snapshot_name)

    # [START howto_operator_rds_start_export_task]
    start_export = RdsStartExportTaskOperator(
        task_id="start_export",
        export_task_identifier=rds_export_task_id,
        source_arn=snapshot_arn,
        s3_bucket_name=bucket_name,
        s3_prefix="rds-test",
        iam_role_arn=test_context[ROLE_ARN_KEY],
        kms_key_id=test_context[KMS_KEY_ID_KEY],
    )
    # [END howto_operator_rds_start_export_task]

    # RdsStartExportTaskOperator waits by default, setting as False to test the Sensor below.
    start_export.wait_for_completion = False

    # [START howto_operator_rds_cancel_export]
    cancel_export = RdsCancelExportTaskOperator(
        task_id="cancel_export",
        export_task_identifier=rds_export_task_id,
    )
    # [END howto_operator_rds_cancel_export]
    cancel_export.check_interval = 10
    cancel_export.max_attempts = 120

    # [START howto_sensor_rds_export_task_existence]
    export_sensor = RdsExportTaskExistenceSensor(
        task_id="export_sensor",
        export_task_identifier=rds_export_task_id,
        target_statuses=["canceled"],
    )
    # [END howto_sensor_rds_export_task_existence]

    delete_snapshot = RdsDeleteDbSnapshotOperator(
        task_id="delete_snapshot",
        db_type="instance",
        db_snapshot_identifier=rds_snapshot_name,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_bucket = S3DeleteBucketOperator(
        task_id="delete_bucket",
        bucket_name=bucket_name,
        force_delete=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_db_instance = RdsDeleteDbInstanceOperator(
        task_id="delete_db_instance",
        db_instance_identifier=rds_instance_name,
        rds_kwargs={"SkipFinalSnapshot": True},
        trigger_rule=TriggerRule.ALL_DONE,
    )

    chain(
        # TEST SETUP
        test_context,
        create_bucket,
        create_db_instance,
        create_snapshot,
        await_snapshot,
        snapshot_arn,
        # TEST BODY
        start_export,
        cancel_export,
        export_sensor,
        # TEST TEARDOWN
        delete_snapshot,
        delete_bucket,
        delete_db_instance,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

from airflow.providers.amazon.aws.operators.sagemaker_unified_studio import (
    SageMakerNotebookOperator,
)

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

"""
Prerequisites: The account which runs this test must manually have the following:
1. An IAM IDC organization set up in the testing region with a user initialized
2. A SageMaker Unified Studio Domain (with default VPC and roles)
3. A project within the SageMaker Unified Studio Domain
4. A notebook (test_notebook.ipynb) placed in the project's s3 path

This test will emulate a DAG run in the shared MWAA environment inside a SageMaker Unified Studio Project.
The setup tasks will set up the project and configure the test runner to emulate an MWAA instance.
Then, the SageMakerNotebookOperator will run a test notebook. This should spin up a SageMaker training job, run the notebook, and exit successfully.
"""

DAG_ID = "example_sagemaker_unified_studio"

# Externally fetched variables:
DOMAIN_ID_KEY = "DOMAIN_ID"
PROJECT_ID_KEY = "PROJECT_ID"
ENVIRONMENT_ID_KEY = "ENVIRONMENT_ID"
S3_PATH_KEY = "S3_PATH"
REGION_NAME_KEY = "REGION_NAME"

sys_test_context_task = (
    SystemTestContextBuilder()
    .add_variable(DOMAIN_ID_KEY)
    .add_variable(PROJECT_ID_KEY)
    .add_variable(ENVIRONMENT_ID_KEY)
    .add_variable(S3_PATH_KEY)
    .add_variable(REGION_NAME_KEY)
    .build()
)


def get_mwaa_environment_params(
    domain_id: str,
    project_id: str,
    environment_id: str,
    s3_path: str,
    region_name: str,
):
    AIRFLOW_PREFIX = "AIRFLOW__WORKFLOWS__"

    parameters = {}
    parameters[f"{AIRFLOW_PREFIX}DATAZONE_DOMAIN_ID"] = domain_id
    parameters[f"{AIRFLOW_PREFIX}DATAZONE_PROJECT_ID"] = project_id
    parameters[f"{AIRFLOW_PREFIX}DATAZONE_ENVIRONMENT_ID"] = environment_id
    parameters[f"{AIRFLOW_PREFIX}DATAZONE_SCOPE_NAME"] = "dev"
    parameters[f"{AIRFLOW_PREFIX}DATAZONE_STAGE"] = "prod"
    parameters[f"{AIRFLOW_PREFIX}DATAZONE_ENDPOINT"] = f"https://datazone.{region_name}.api.aws"
    parameters[f"{AIRFLOW_PREFIX}PROJECT_S3_PATH"] = s3_path
    parameters[f"{AIRFLOW_PREFIX}DATAZONE_DOMAIN_REGION"] = region_name
    return parameters


@task
def mock_mwaa_environment(parameters: dict):
    """
    Sets several environment variables in the container to emulate an MWAA environment provisioned
    within SageMaker Unified Studio. When running in the ECSExecutor, this is a no-op.
    """
    import os

    for key, value in parameters.items():
        os.environ[key] = value


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()

    test_env_id = test_context[ENV_ID_KEY]
    domain_id = test_context[DOMAIN_ID_KEY]
    project_id = test_context[PROJECT_ID_KEY]
    environment_id = test_context[ENVIRONMENT_ID_KEY]
    s3_path = test_context[S3_PATH_KEY]
    region_name = test_context[REGION_NAME_KEY]

    mock_mwaa_environment_params = get_mwaa_environment_params(
        domain_id,
        project_id,
        environment_id,
        s3_path,
        region_name,
    )

    setup_mwaa_environment = mock_mwaa_environment(mock_mwaa_environment_params)

    # [START howto_operator_sagemaker_unified_studio_notebook]
    notebook_path = "test_notebook.ipynb"  # This should be the path to your .ipynb, .sqlnb, or .vetl file in your project.

    run_notebook = SageMakerNotebookOperator(
        task_id="run-notebook",
        input_config={"input_path": notebook_path, "input_params": {}},
        output_config={"output_formats": ["NOTEBOOK"]},  # optional
        compute={
            "instance_type": "ml.m5.large",
            "volume_size_in_gb": 30,
        },  # optional
        termination_condition={"max_runtime_in_seconds": 600},  # optional
        tags={},  # optional
        wait_for_completion=True,  # optional
        waiter_delay=5,  # optional
        deferrable=False,  # optional
        executor_config={  # optional
            "overrides": {
                "containerOverrides": [
                    {
                        "environment": [
                            {"name": key, "value": value}
                            for key, value in mock_mwaa_environment_params.items()
                        ],
                        "name": "ECSExecutorContainer",  # Necessary parameter
                    }
                ]
            }
        },
    )
    # [END howto_operator_sagemaker_unified_studio_notebook]

    chain(
        # TEST SETUP
        test_context,
        setup_mwaa_environment,
        # TEST BODY
        run_notebook,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from airflow.providers.amazon.aws.hooks.ec2 import EC2Hook
from airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor
from airflow.providers.amazon.aws.triggers.ec2 import EC2StateSensorTrigger
from airflow.providers.amazon.aws.utils import validate_execute_complete_event
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.common.compat.sdk import AirflowException, conf

if TYPE_CHECKING:
    from airflow.sdk import Context


class EC2InstanceStateSensor(AwsBaseSensor[EC2Hook]):
    """
    Poll the state of the AWS EC2 instance until the instance reaches the target state.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:EC2InstanceStateSensor`

    :param target_state: target state of instance
    :param instance_id: id of the AWS EC2 instance
    :param region_name: (optional) aws region name associated with the client
    :param deferrable: if True, the sensor will run in deferrable mode
    """

    aws_hook_class = EC2Hook
    template_fields: Sequence[str] = aws_template_fields("target_state", "instance_id", "region_name")
    ui_color = "#cc8811"
    ui_fgcolor = "#ffffff"
    valid_states = ["running", "stopped", "terminated"]

    def __init__(
        self,
        *,
        target_state: str,
        instance_id: str,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ):
        if target_state not in self.valid_states:
            raise ValueError(f"Invalid target_state: {target_state}")
        super().__init__(**kwargs)
        self.target_state = target_state
        self.instance_id = instance_id
        self.deferrable = deferrable

    def execute(self, context: Context) -> Any:
        if self.deferrable:
            self.defer(
                trigger=EC2StateSensorTrigger(
                    instance_id=self.instance_id,
                    target_state=self.target_state,
                    aws_conn_id=self.aws_conn_id,
                    region_name=self.region_name,
                    poll_interval=int(self.poke_interval),
                ),
                method_name="execute_complete",
            )
        else:
            super().execute(context=context)

    def poke(self, context: Context):
        instance_state = self.hook.get_instance_state(instance_id=self.instance_id)
        self.log.info("instance state: %s", instance_state)
        return instance_state == self.target_state

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> None:
        validated_event = validate_execute_complete_event(event)

        if validated_event["status"] != "success":
            raise AirflowException(f"Error: {validated_event}")
